//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                              function module                               *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// Built in functions able to be called from Z80 code.
//
// General purpose functions that don't belong to any particular module.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.1.0 - 22 June 2009, S.J.Kay
// - Fixed a Big Endian issue. 'leu16_to_host(ubee_addr | (data << 8))'
//   should have been just 'ubee_addr | (data << 8)' in function_ubee_w()
// - Made improvements to the logging process.
//
// v4.0.0 - 7 June 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Changes made to access the current Z80 DRAM/SRAM memory as the memory
//   management method has completely changed.
// - More Endian issues found and corrected (needed for Big Endian HW)
// - Changed MD5 file matching to keep trying other matching MD5 entries when
//   the file name is not found.
// - Masked off appended register values (upper 8 bits) from port values where
//   these were required.
//
// v3.1.0 - 20 April 2009, S.J.Kay
// - Fixed swap_endian() function, was doing nothing!
// - Added create_md5() function.
// - Added optional MD5 lookup for ROMs and disks in the 'roms.alias' and
//   'disks.alias' files with new function find_file_alias().
// - Moved functions function_emu_command(), function_repeat_start(),
//   function_repeat_stop() and function_update() to keyb_*() in keyb.c.
// - Added time_get_secs() function to return timer ticks in seconds.
// - Changed all printf() calls to use a local xprintf() function.
// - Added EMU_CMD_CONSOLE function to enter console mode.
// - Changes to stop OpenGL functions being called when using SDL video mode.
// - Added OpenGL conditional code compilation using USE_OPENGL.
//
// v3.0.0 - 10 October 2008, S.J.Kay
// - Changed file_readline() function to remove leading spaces.
// - Added EMU_CMD_MWHEEL function to select mouse wheel association.
// - Added EMU_CMD_SCREENI and EMU_CMD_SCREEND  to function_emu_command()
// - Added EMU_CMD_VIDSIZE2 function to set screen size by set amount.
// - Added EMU_CMD_GL_FILTER toggling of OpenGL filter mode.
// - Changed the gui_toggledisplay() function to video_toggledisplay() as
//   code has moved to a new video.c module.
// - Added time_get_ms() function and removed repeated timer code.
// - Added time_delay_ms() function.
//
// v2.8.0 - 1 September 2008, S.J.Kay
// - Added new functions function_repeat_start(), function_repeat_stop() and
//   function_update().  These are used to make repeated calls to the
//   function_emu_command() function.
// - Added EMU_CMD_VOLUMEI and EMU_CMD_VOLUMED to function_emu_command()
// - Changes to the way gui_status_update() is called.
// - Added EMU_CMD_PAUSE to function_emu_command function.  Debug modes
//   modified to not activate while in a paused state.
// - Added function 0x06 to return the --file-list string count.
// - Added function 0x10 to set the current --file-list string position.
// - Modified function 0x00 function to return the current string for the
//   --file-list option parameter string. Compatible with earlier applications
//   that made use of this function.
// - Cleaned up some code that was producing 2 warnings.
//
// v2.7.0 - 27 June 2008, S.J.Kay
// - Added function_basic() functions.
// - Added a function_application() functions.
// - Added find_file_entry() function to find what file to use for file name
//   aliases defined in a configuration file.
// - Moved options_readline() function in options.c module over and is now
//   file_readline() function, with file pointer, and string size passed.
// - Added open_file() function.  Opens/creates/tests files, this also
//   provides alternate path checking and allows slash characters to be
//   win32 or unix style.
// - Added structure emu_t and moved variables into it.
// - Added structure func_t and moved variables into it.
// - Added EMU_CMD_MUTE to function_emu_command() function.
// - Added functions to function_control() to enable video and sound.
// - Added sound_t and crtc_t structure.
//
// v2.6.0 - 11 May 2008, S.J.Kay
// - Added a function_joystick() functions.
// - Added toupper_string() function.
// - Moved the string_search() function over from the options.c module.
// - Fixed an endianess issue in function_stdio() function for the 'putchar
//   function'. 
// - File name matching in win32 fixed in function "case 0x02 return the next
//   directory entry (with wildcards)". Was only able to match file names
//   that used all lower case. CP/M support program HOST2CPM should now work.
//
// v2.3.0 - 23 January 2007, S.J.Kay
// - Added function_emu_command() function.
// - Added printf, hex and decimal output functions to the stdio group.
// - Added dump function to diagnostics group.
// - Added additional functions to status group.
// - Some function values that are intended to return 0 or 1 were missing
//   the little_endian conversion, so 1 would have returned as some other non
//   zero value on big endian platforms. If the test in an application was for
//   a non zero value then there won't be a problem.  Constant values of 0 now
//   also use the little_endian conversion for consistency.
// - Added modio_t debug_t and regdump_t structures.
//
// v2.2.0 - 11 December 2007, S.J.Kay
// - Added status, control, keyboard, stdio, and file and directory functions.
// - Changed the method for Application and Version reporting.
// - Added a Z80 memory dump function to stdio.
//
// v1.4.0 - 22 September 2007, S.J.Kay
// - Added copy_file() function.
//
// v1.2.0 - 19 August 2007, S.J.Kay
// - Created new module to implement the emulator function ports.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <sys/stat.h>

#ifdef MINGW
#include <windows.h>
#else
#include <sys/types.h> 		// various type definitions, like pid_t
#include <sys/time.h>
#endif

#include "ubee512.h"
#include "function.h"
#include "z80debug.h"
#include "gui.h"
#include "crtc.h"
#include "rtc.h"
#include "sound.h"
#include "joystick.h"
#include "console.h"
#include "z80.h"
#include "z80api.h"
#include "md5.h"
#include "memmap.h"

//==============================================================================
// structures and variables
//
// NOTE:
// - All Z80 structure members are in Little Endian format.  This requires
//   conversion to and from the host Endian format.  The exception is:
//
//   f->file.fp.d
//
//   This is a file pointer and is used in the host's native format.
//
// - Access to memory locations that cross a 0x8000 Z80 memory bank will not
//   work except for fread and fwrite functions which use local buffering
//   and copying data to/from the correct banks.
//==============================================================================
func_t func =
{
 .file_exit = 1
};

#ifdef MINGW
#else
 static struct timeval tod;
#endif

static const char emulator_id_str[] = APPIDSTR;  // defined in the Makefile
static const char emulator_ver_str[] = APPVER;   // defined in the Makefile
static int emulator_id_pos;

static int ubee_req;
static int ubee_addr;
static int ubee_command;

static char *z80mem_f;
static char *z80mem_x1;
static char *z80mem_x2;
static char *z80mem_x3;
static char *z80mem_x4;

extern char userhome_confpath[];
extern char userhome_romspath[];
extern char userhome_diskpath[];

extern emu_t emu;
extern const model_t model_data[];
extern model_t modelx;
extern modio_t modio;
extern debug_t debug;
extern regdump_t regdump;
extern joystick_t joystick;
extern crtc_t crtc;
extern sound_t sound;

//==============================================================================
// Get the current clock time in seconds
//
//   pass: void
// return: int				number of seconds
//==============================================================================
int time_get_secs (void)
{
 return time_get_ms() / 1000;
}

//==============================================================================
// Get the current clock time in milliseconds
//
//   pass: void
// return: uint64_t			number of milliseconds
//==============================================================================
uint64_t time_get_ms (void)
{
#ifdef MINGW
 return (uint64_t)clock();
#else
 gettimeofday(&tod, NULL);          
 return (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif
}

//==============================================================================
// Time delay in milliseconds. Gives up host CPU time to other applications.
//
//   pass: int ms			number of milliseconds
// return: void
//==============================================================================
void time_delay_ms (int ms)
{
 SDL_Delay(ms);
}

//==============================================================================
// Time wait in milliseconds. Does NOT give up host CPU time!
//
//   pass: int ms			number of milliseconds
// return: void
//==============================================================================
void time_wait_ms (int ms)
{
 int t = time_get_ms() + ms;
 while (time_get_ms() < t)
    ;
}

//==============================================================================
// Get the z80 map pointer and return the address masked with 0x7fff
//
//   pass: int *addr
// return: char *
//==============================================================================
char *get_z80mem_ptr_and_addr (int *addr)
{
 char *map = memmap_get_z80_ptr(*addr);
 *addr &= 0x7fff;
 return map;
}

//==============================================================================
// Copy a string to upper case.  Source and destination may be the same.
//
//   pass: char *dest			destination string
//         char *src			source string
// return: void
//==============================================================================
void toupper_string (char *dest, char *src)
{
 int i = 0;
 
 while (src[i])
    {
     dest[i] = toupper(src[i]);
     i++;
    }
 dest[i] = 0;
}

//==============================================================================
// Read a text line from a file.
//
// - Leading spaces are removed.
// - Comment and empty lines are not returned.
// - Control characters and space characters are removed from the end of the
//   line.
// - An empty string is returned if an error occurs or the end of the
//   file is reached.
//
//   pass: FILE *fp
//         char *s
//         int size
// return: int				string length
//==============================================================================
int file_readline (FILE *fp, char *s, int size)
{
 int l;
 int i;
 char *res;
 char temp_str[5000];

 do
    {
     res = fgets(temp_str, sizeof(temp_str)-1, fp);
     if (res)
        {
         // remove leading spaces
         i = 0;
         while ((temp_str[i]) && (temp_str[i] <= ' '))
            i++;
         strncpy(s, &temp_str[i], size);
         s[size-1] = 0;

         // check if this is a commented out line and ignore if is
         if ((s[0] == '#') || (s[0] == ';'))
            {
             s[0] = 0;
             l = 0;
            }
         else
            {
             // otherwise remove trailing control and space characters
             l = strlen(s);
             while ((l) && (s[l-1] <= ' '))
                l--;
             s[l] = 0;   
            }    
        }
     else
        {
         s[0] = 0;
         l = 0;
        } 
    }        
 while ((res) && (! l));
 return l;
}

//==============================================================================
// Copy a file from one location to another.
//
//   pass: char *dest			destination file path and name
//         char *src			source file path and name
// return: int				0 if no error, else -1
//==============================================================================
int copy_file (char *dest, char *src)
{
 FILE *srcfp;
 FILE *destfp; 
 char *buf[0x4000];
 int len;

 srcfp = fopen(src, "rb");
 if (srcfp == NULL)
    return -1;

 destfp = fopen(dest, "wb"); 
 if (destfp == NULL)
    {
     fclose(srcfp);		// close the source file
     return -1;
    }

 while (! feof(srcfp))
    {
     len = fread(buf, 1, sizeof(buf), srcfp);
     if (len)
        fwrite(buf, 1, len, destfp);
    }

 fclose(srcfp);
 fclose(destfp); 

 return 0;
}

//==============================================================================
// Convert path slash characters to host format
//
//   pass: char *path		path to be converted
// return: void
//==============================================================================
void convert_slash (char *path)
{
 int i = 0;

// convert slash characters to the host's file format if enabled.
 if (emu.slashconv)
    {
     while (path[i])
        {
         if (path[i] == SLASHCHAR_OTHER)
            path[i] = SLASHCHAR;
         i++;
        }
    }   
}

//==============================================================================
// Open, create or test a file and return the full path.
//
// Mode does not contain 'w'
// -------------------------
// When opening an existing file a test is done to see if the file path
// exists, if not then checks the default location.  If mode is empty ("")
// then only test if the file exists using "rb" as the mode.
//
// Modes:
// "r"   Open a text file for reading
// "a"   Append to a text file
// "rb"  Open a binary file for reading
// "ab"  Append to a binary file
// "r+"  Open a text file for read/write
// "a+"  Open a text file for read/write
// "rb+" Open a binary file for read/write
// "ab+" Open a binary file for read/write    
//
// Mode contains 'w'
// -----------------
// If a file is to be created then the default path is prepended if no
// path slash characters are found.
//
// Modes:
// "w"   Create a text file for writing
// "wb"  Create a binary file for writing
// "w+"  Create a text file for read/write
// "wb+" Create a binary file for read/write
//
// If emu.slashconv is true then slash slash characters are converted to the
// host system format so either '\' or '/' may be used with the same
// results.
//
//   pass: char *path1		path to be tested
//         char *path2          alternate path prefix to be tried
//         char *path3		the full path
//         char *mode           how to open the file, or "" if test only
//         
// return: *FILE		NULL if no path or file could not be created
//==============================================================================
FILE *open_file (const char *path1, const char *path2, char *path3, const char *mode)
{
 int uses_path1 = 0;
 char pathx[SSIZE1];
 FILE *fp;
 
 strcpy(pathx, path1);

// convert slash characters to the host's file format if enabled.
 convert_slash(pathx);

// if a '/', './' or '../' characters are found at the start then treat as
// the path to be used, the default path (path2) will not be prepened in
// that case.
   if ((strstr(pathx, "."SLASHCHAR_STR) == pathx) ||
       (strstr(pathx, ".."SLASHCHAR_STR) == pathx) ||
       (strstr(pathx, SLASHCHAR_STR) == pathx))
      uses_path1 = 1;
      
// if Win32 then a ':' character will also be treated as a path to be used.
#ifdef MINGW
   if (strchr(pathx, ':'))
      uses_path1 = 1;
#endif    

 strcpy(path3, pathx);
 
 if (strchr(mode, 'w'))
// a file is to be created
    {
     if (! uses_path1)
        snprintf(path3, SSIZE1, "%s%s", path2, pathx);
     fp = fopen(path3, mode);
     return fp;
    }

// test if a file exists, if uses_path1 is true then only the current
// directory will be checked.
 if (! strlen(mode))
    {
     fp = fopen(path3, "rb");
     if (fp)
        {
         fclose(fp);
         return fp;
        }

     if (uses_path1)
        return fp;

     snprintf(path3, SSIZE1, "%s%s", path2, pathx);
     fp = fopen(path3, "rb");
     if (fp)
        fclose(fp);
     return fp;    
    }

// open an existing file, if uses_path1 is true then only the current
// directory will be checked.
 fp = fopen(path3, mode);
 if (fp || uses_path1)
    return fp;

 snprintf(path3, SSIZE1, "%s%s", path2, pathx);
 fp = fopen(path3, mode);
 return fp;
}

//==============================================================================
// Test if a file exists and return the full path.
//
//   pass: char *path1		path to be tested
//         char *path2          alternate path prefix to be tried
//         char *path3		the full path that was used
//         
// return: *FILE		NULL if no path or file found
//==============================================================================
FILE *test_file (const char *path1, const char *path2, char *path3)
{
 return open_file(path1, path2, path3, "");
}

//==============================================================================
// Search an array of strings for the first occurence of the passed search
// string.  The string array must be be terminated by an empty string. 
//
//   pass: char *strg_array	pointer to a string array to be checked
//         char *strg_find	pointer to a string to search for
// return: int			index if found, else -1
//==============================================================================
int string_search (char *strg_array[], char *strg_find)
{
 int i;
 
 for (i=0; strg_array[i][0] != 0; i++)
    {
     if (strcmp(strg_find, strg_array[i]) == 0)
        return i;
    }
 return -1;
}

//==============================================================================
// Open a text file and search for a non case sensitive matching string
// entry in the first column, if found then return the entry from the second
// column.  If the second column is empty then a null terminated string is
// returned.
//
// Lines beginning with '#' or ';' are ignored.
//
//   pass: char *filename	file name
//         char *strg_search	string to be checked (1st column)
//         char *strg_value	string returned (2nd column)
// return: int			1 if entry matched and has a value, else 0
//==============================================================================
static int find_file_entry (char *filename, char *strg_search, char *strg_value)
{
 FILE *fp;
 char filepath[SSIZE1];
 char s[1000];
 int i;
 
 strg_value[0] = 0;
 fp = open_file(filename, userhome_confpath, filepath, "r");
 if (! fp)
    return 0;

 while (1)
    {
     if (file_readline(fp, s, sizeof(s)))
        {
         if (strcasestr(s, strg_search) == s)
            {
             fclose(fp);
             i = 0;
             while (s[i] > ' ')
                i++;
             while (s[i] <= ' ')
                i++;
             if (i < strlen(s))
                {
                 strncpy(strg_value, &s[i], SSIZE1);
                 strg_value[SSIZE1-1] = 0;
                 return 1;
                } 
             return 0;   
            } 
        }
     else
        {
         fclose(fp);
         return 0;
        } 
    }
}

//==============================================================================
// Open a text file and search for a non case sensitive matching string
// entry in the first column, if found then check if the file in the second
// column exists, if not then repeat until another matching MD5 and file is
// found or the end of file is reached.
//
// Lines beginning with '#' or ';' are ignored.
//
//   pass: char *filename	MD5 file to be used
//         char *strg_search	MD5 value to be searched for (1st column)
//         char *strg_value	file name returned (2nd column)
// return: int			1 if entry matched and file existed, else 0
//==============================================================================
static int find_md5_file_entry (char *filename, char *strg_search, char *strg_value)
{
 FILE *fp;
 char filepath[SSIZE1];
 char s[1000];
 int i;
 int found;
 
 strg_value[0] = 0;
 fp = open_file(filename, userhome_confpath, filepath, "r");
 if (! fp)
    return 0;

 while (1)
    {
     if (file_readline(fp, s, sizeof(s)))
        {
         if (strcasestr(s, strg_search) == s)
            {
             i = 0;
             while (s[i] > ' ')
                i++;
             while (s[i] <= ' ')
                i++;
        
             if (i < strlen(s))
                {
                 strncpy(strg_value, &s[i], SSIZE1);
                 strg_value[SSIZE1-1] = 0;

                 // if the file was calculated as a binary MD5 remove the leading '*'
                 if (strg_value[0] == '*')
                    strcpy(strg_value, &strg_value[1]);
                 
                 // check if a file exists for this MD5 entry
                 if (strcmp(filename, "roms.md5") == 0)
                    found = (test_file(strg_value, userhome_romspath, filepath) != NULL);
                 else
                    found = (test_file(strg_value, userhome_diskpath, filepath) != NULL);
                    
                 if (emu.verbose)
                    {
                     xprintf("%s=%s %s", filename, strg_search, strg_value);
                     if (found)
                        xprintf("\n");
                     else   
                        xprintf(" (no file available)\n");
                    }
                    
                 if (found)
                    {
                     fclose(fp);
                     return 1;
                    }
                }
            }
        }
     else
        {
         fclose(fp);
         return 0;
        }
    }
}

//==============================================================================
// Search a text file for a matching string entry in the first column, if
// found then return the entry from the second column or if the value is
// prefixed with 'md5=' use the MD5 value in a second search of the
// 'roms.md5' or 'disks.md5' file for a corresponding filename.  If the
// second column is empty then a null terminated string is returned.
//
// Lines beginning with '#' or ';' are ignored.
//
//   pass: char *filename	file name
//         char *strg_search	string to be checked (1st column)
//         char *strg_value	string returned (2nd column)
// return: int			1 if entry matched and has a value, else 0
//==============================================================================
int find_file_alias (char *filename, char *strg_search, char *strg_value)
{
 char *p;
 char md5_filename[20];

 if (find_file_entry(filename, strg_search, strg_value))
    {
     // if an MD5 is present then find the MD5 file
     p = strcasestr(strg_value, "md5=");
     if (p == strg_value)
        {
         strcpy(strg_search, p+4);
   
         if (strcmp(filename, ALIASES_ROMS) == 0)
            strcpy(md5_filename, "roms.md5");
         else
            if (strcmp(filename, ALIASES_DISKS) == 0)
               strcpy(md5_filename, "disks.md5");
            else
               return 1;
   
         if (find_md5_file_entry(md5_filename, strg_search, strg_value))
            return 1;
         else
            return 0;
        }
     else
        return 1;   // have a non-MD5 alias entry
    }
 else
    return 0;
}

//==============================================================================
// Create an MD5 message digest.
//
//   pass: char *filename		file to create MD5 for
//         char *md5			pointer to a 33 byte buffer for MD5
// return: void
//==============================================================================
void create_md5 (char *filename, char *md5)
{
 FILE *fp;
 uint8_t resblock[16];
 int i;

 md5[0] = 0;
 fp = fopen(filename, "rb");
 if (fp)
    {
     if (md5_stream(fp, &resblock) == 0)
        {
         for (i = 0; i < 16; i++)
            sprintf(md5, "%s%02x", md5, resblock[i]);
        } 
     fclose(fp);        
    } 
}

#ifdef MINGW
//==============================================================================
// strcasestr for systems that don't have this function.
//
//   pass: char *haystack
//         char *needle
// return: void
//==============================================================================
char *strcasestr (char *haystack, char *needle)
{
 char *p, *startn = 0, *np = 0;
 
 for (p = haystack; *p; p++)
    {
     if (np)
        {
         if (toupper(*p) == toupper(*np))
            {
             if (!*++np)
                return startn;
            }
         else
            np = 0;
        }
     else
        if (toupper(*p) == toupper(*needle))
           {
            np = needle + 1;
            startn = p;
           }
    }
 return 0;
}

//==============================================================================
// strcasecmp for systems that don't have this function.
//
//   pass: const char *s1
//         const char *s2
// return: void
//==============================================================================
int strcasecmp (const char *s1, const char *s2)
{
 if (!s1 && !s2)
    return 0;
    
 if (!s1)
    return 1;
    
 if (!s2)
    return -1;
    
 while (*s1 && *s2 && tolower((unsigned char)*s1) == tolower((unsigned char)*s2))
    {
     s1++;
     s2++;
    }
    
 return tolower((unsigned char)*s1) - tolower((unsigned char)*s2);
}
#endif

//==============================================================================
// Compare strings while treating digits characters numerically.
//
// Compare S1 and S2 as strings holding indices/version numbers, returning
// less than, equal to or greater than zero if S1 is less than, equal to or
// greater than S2 (for more info, see the texinfo doc).
//
// This code has been included as mingw does not appear to have a function
// to do this.
//
// Changed the function name to xstrverscmp() as strverscmp() is an illegal
// use of 'str' for the start of a non 'C' library function name.
//
// Copyright (C) 1997, 2000, 2002, 2004, 2006 Free Software Foundation, Inc.
// This function is part of the GNU C Library.  Contributed by Jean-François
// Bignolles <bignolle@ecoledoc.ibp.fr>, 1997.
//
//   pass: const char *s1
//         const char *s2
// return: int
//==============================================================================
// states: S_N: normal, S_I: comparing integral part, S_F: comparing
//           fractional parts, S_Z: idem but with leading Zeroes only

#define S_N    0x0
#define S_I    0x4
#define S_F    0x8
#define S_Z    0xC

// result_type: CMP: return diff; LEN: compare using len_diff/diff
#define CMP    2
#define LEN    3

// ISDIGIT differs from isdigit, as follows:
//  - Its arg may be any int or unsigned int; it need not be an unsigned char
//    or EOF.
//  - It's typically faster.
//  POSIX says that only '0' through '9' are digits.  Prefer ISDIGIT to
//  isdigit unless it's important to use the locale's definition
//  of `digit' even when the host does not conform to POSIX.

#define ISDIGIT(c) ((unsigned int) (c) - '0' <= 9)

int xstrverscmp (const char *s1, const char *s2)
{
 const unsigned char *p1 = (const unsigned char *) s1;
 const unsigned char *p2 = (const unsigned char *) s2;
 unsigned char c1, c2;
 int state;
 int diff;

 // Symbol(s)    0       [1-9]   others  (padding)
 // Transition   (10) 0  (01) d  (00) x  (11) -
 static const unsigned int next_state[] =
 {
     /* state    x    d    0    - */
     /* S_N */  S_N, S_I, S_Z, S_N,
     /* S_I */  S_N, S_I, S_I, S_I,
     /* S_F */  S_N, S_F, S_F, S_F,
     /* S_Z */  S_N, S_F, S_Z, S_Z
 };

 static const int result_type[] =
 {
     /* state   x/x  x/d  x/0  x/-  d/x  d/d  d/0  d/-
                0/x  0/d  0/0  0/-  -/x  -/d  -/0  -/- */

     /* S_N */  CMP, CMP, CMP, CMP, CMP, LEN, CMP, CMP,
                CMP, CMP, CMP, CMP, CMP, CMP, CMP, CMP,
     /* S_I */  CMP, -1,  -1,  CMP,  1,  LEN, LEN, CMP,
                 1,  LEN, LEN, CMP, CMP, CMP, CMP, CMP,
     /* S_F */  CMP, CMP, CMP, CMP, CMP, LEN, CMP, CMP,
                CMP, CMP, CMP, CMP, CMP, CMP, CMP, CMP,
     /* S_Z */  CMP,  1,   1,  CMP, -1,  CMP, CMP, CMP,
                -1,  CMP, CMP, CMP
 };

 if (p1 == p2)
    return 0;

 c1 = *p1++;
 c2 = *p2++;

 // Hint: '0' is a digit too.
 state = S_N | ((c1 == '0') + (ISDIGIT (c1) != 0));

 while ((diff = c1 - c2) == 0 && c1 != '\0')
    {
     state = next_state[state];
     c1 = *p1++;
     c2 = *p2++;
     state |= (c1 == '0') + (ISDIGIT (c1) != 0);
    }

 state = result_type[state << 2 | ((c2 == '0') + (ISDIGIT (c2) != 0))];

 switch (state)
    {
     case CMP:
        return diff;

        case LEN:
           while (ISDIGIT (*p1++))
              if (!ISDIGIT (*p2++))
                 return 1;

           return ISDIGIT (*p2) ? -1 : diff;

        default:
           return state;
    }
}

//==============================================================================
// Test if system uses Big Endian or Little Endian format.
//
//   pass: void
// return: int				1 if Big Endian, 0 if Little Endian
//==============================================================================
static int IsBigEndian (void)
{
#if defined(__aix__)
 return 1;
#else
 int i = 1;
 int c = *((char*)&i);
 return c == 0;
#endif
}

//==============================================================================
// Swap Endianess. Only works for even number variable sizes.
//
//   pass: unsigned char *na
//         int size			size in bytes of the variable
// return: void
//==============================================================================
void swap_endian (unsigned char *na, int size) 
{
 int i;
 int x;
 unsigned char *nb = na + (size - 1);  // pointer to end of variable
 
 for (i = 0; i < (size / 2); i++)
    {
     x = *na;
     *na++ = *nb;
     *nb-- = x;
    } 
}

//==============================================================================
// Little Endian to host endian,  and host endian to Little Endian functions.
// Swaps byte order if on a Big Endian host.
//
// Don't use these functions directly.  use the macros defined in the header
// file as these make the usage clearer.
//
//   pass: xintxx_t n			little endian, or host endian format
// return: xintxx_t			little endian format
//==============================================================================
int16_t little_endian_16 (int16_t n)
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

uint16_t little_endian_u16 (uint16_t n)
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

int32_t little_endian_32 (int32_t n) 
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

uint32_t little_endian_u32 (uint32_t n)
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

int64_t little_endian_64 (int64_t n) 
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

uint64_t little_endian_u64 (uint64_t n) 
{
 if (IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// Big Endian to host endian,  and host to Big Endian functions.
// Swaps byte order if on a Little Endian host.
//
//   pass: xintxx_t n			big endian, or host endian format
// return: xintxx_t			host endian, or big endian format
//==============================================================================
int16_t big_endian_16 (int16_t n)
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}
          
uint16_t big_endian_u16 (uint16_t n)
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);    
}

int32_t big_endian_32 (int32_t n) 
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}
                                              
uint32_t big_endian_u32 (uint32_t n)
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

int64_t big_endian_64 (int64_t n) 
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

uint64_t big_endian_u64 (uint64_t n) 
{
 if (! IsBigEndian())
    swap_endian((unsigned char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// set Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format.  If host uses big endian format the values are swapped first.
//
//   pass: int addr
//         xintxx_t  n
// return: void
//==============================================================================
#if 0
// Note: All these will need z80mem replaced using new access methods.
static void z80_set_8 (int addr, int8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_u8 (int addr, uint8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_16 (int addr, int16_t n)
{
 n = host_to_le16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_u16 (int addr, uint16_t n)
{
 n = host_to_leu16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_32 (int addr, int32_t n)
{
 n = host_to_le32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_u32 (int addr, uint32_t n)
{
 n = host_to_leu32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_64 (int addr, int64_t n)
{
 n = host_to_le64(n);
 memcpy(z80mem+addr, &n, 8);
}

static void z80_set_u64 (int addr, uint64_t n)
{
 n = host_to_leu64(n);
 memcpy(z80mem+addr, &n, 8);
}
#endif

//==============================================================================
// get Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format and convert to host endian format.
//
//   pass: int addr
// return: xintxx_t 			in host endian format
//==============================================================================
#if 0
// Note: All these will need z80mem replaced using new access methods.
static int8_t z80_get_8 (int addr)
{
 return z80mem[addr];
}

static uint8_t z80_get_u8 (int addr)
{
 return z80mem[addr];
}

static int16_t z80_get_16 (int addr)
{
 int16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (le16_to_host(n));
}

static uint16_t z80_get_u16 (int addr)
{
 uint16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (leu16_to_host(n));
}

static int32_t z80_get_32 (int addr)
{
 int32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (le32_to_host(n));
}

static uint32_t z80_get_u32 (int addr)
{
 uint32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (leu32_to_host(n));
}

static int64_t z80_get_64 (int addr)
{
 int64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (le64_to_host(n));
}

static uint64_t z80_get_u64 (int addr)
{
 uint64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (leu64_to_host(n));
}
#endif

//==============================================================================
// Wildcards
//
// Source obtained from GPLv2 wildcards v1.2 project.
// Copyright (C) 1996, 1997, 1998, 1999, 2000 Florian Schintke.
//
// Only structural changes have been made to keep inline with the rest of
// the emulator's style.
//
// int set (char **wildcard, char **test)
// Scans a set of characters and returns 0 if the set mismatches at this
// position in the teststring and 1 if it is matching wildcard is set to the
// closing ] and test is unmodified if mismatched and otherwise the char
// pointer is pointing to the next character
//
// int asterisk (char **wildcard, char **test)
// scans an asterisk
//
// int wildcardfit (char *wildcard, char *test)
// The function to call to compare wildcards to the test string.
//
//   pass: char *wildcard
//         char *test
// return: int			1 if wildcards fit, else 0
//==============================================================================
int asterisk (char **wildcard, char **test);
int wildcardfit (char *wildcard, char *test);

int set (char **wildcard, char **test)
{
 int fit = 0;
 int negation = 0;
 int at_beginning = 1;

 if ('!' == **wildcard)
    {
     negation = 1;
     (*wildcard)++;
    }
 while ((']' != **wildcard) || (1 == at_beginning))
    {
     if (0 == fit)
        {
         if (('-' == **wildcard) 
            && ((*(*wildcard - 1)) < (*(*wildcard + 1)))
            && (']' != *(*wildcard + 1))
	    && (0 == at_beginning))
            {
             if (((**test) >= (*(*wildcard - 1))) && ((**test) <= (*(*wildcard + 1))))
                {
                 fit = 1;
                 (*wildcard)++;
                }
            }
         else
            if ((**wildcard) == (**test))
               {
                fit = 1;
               }
        }
     (*wildcard)++;
     at_beginning = 0;
    }
 if (1 == negation)
// change from zero to one and vice versa
 fit = 1 - fit;
 if (1 == fit) 
    (*test)++;
 return (fit);
}

int asterisk (char **wildcard, char **test)
{
// Warning: uses multiple returns
 int fit = 1;

// erase the leading asterisk
 (*wildcard)++; 
 while (('\000' != (**test)) && (('?' == **wildcard) || ('*' == **wildcard)))
    {
     if ('?' == **wildcard) 
	(*test)++;
     (*wildcard)++;
    }

// Now it could be that test is empty and wildcard contains
// aterisks. Then we delete them to get a proper state
 while ('*' == (**wildcard))
    (*wildcard)++;

 if (('\0' == (**test)) && ('\0' != (**wildcard)))
    return (fit = 0);
 if (('\0' == (**test)) && ('\0' == (**wildcard)))
    return (fit = 1); 
 else
    {
     // Neither test nor wildcard are empty!
     // the first character of wildcard isn't in [*?]
     if (0 == wildcardfit(*wildcard, (*test)))
	{
         do 
	    {
	     (*test)++;
	     // skip as much characters as possible in the teststring
	     // stop if a character match occurs
	      while (((**wildcard) != (**test)) && ('['  != (**wildcard)) && ('\0' != (**test)))
		 (*test)++;
	    }
	 while ((('\0' != **test))? (0 == wildcardfit (*wildcard, (*test))) : (0 != (fit = 0)));
	}
     if (('\0' == **test) && ('\0' == **wildcard))
	fit = 1;
     return (fit);
    }
}

int wildcardfit (char *wildcard, char *test)
{
 int fit = 1;
  
 for (; ('\000' != *wildcard) && (1 == fit) && ('\000' != *test); wildcard++)
    {
     switch (*wildcard)
        {
         case '[': wildcard++; // leave out the opening square bracket
                   fit = set (&wildcard, &test);
                   // we don't need to decrement the wildcard as in case
                   // of asterisk because the closing ] is still there
                   break;
         case '?': test++;
                   break;
         case '*': fit = asterisk(&wildcard, &test);
                   // the asterisk was skipped by asterisk() but the loop will
                   // increment by itself. So we have to decrement
                   wildcard--;
                   break;
          default: fit = (int) (*wildcard == *test);
                   test++;
        }
    }
 while ((*wildcard == '*') && (1 == fit)) 
// here the teststring is empty otherwise you cannot
// leave the previous loop
 wildcard++;
 return (int) ((1 == fit) && ('\0' == *test) && ('\0' == *wildcard));
}
 
//==============================================================================
// Function Initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_init (void)
{
 ubee_req = 0;
 return 0;
}

//==============================================================================
// Function de-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_deinit (void)
{
 return 0;
}

//==============================================================================
// Function reset
//
//   pass: void
// return: int			0
//==============================================================================
int function_reset (void)
{
 ubee_req = 0;
 return 0;
}

//==============================================================================
// uBee512 read - Special Ubee512 Port function
//
// This function allows a Z80 program to identify that it is running in the
// uBee512 emulator.  This is provided so that some extra functionality can
// be taken advantage off and must be called and checked before using any
// special uBee512 write function calls to prevent unknown results on a real
// Microbee.
//
// This port call just returns the program name as a string one character at
// a time in a circular loop, the APPIDSTR in the Makefile ultimately
// determines the name to be used.
//
// Version information can be returned by a using a status function call.
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: uint16_t		returns a null terminated string in a loop
//==============================================================================
uint16_t function_ubee_r (uint16_t port, struct z80_port_read *port_s)
{
 char c;

 c = emulator_id_str[emulator_id_pos++];
 if (! c)
    emulator_id_pos = 0;
 
 if (modio.func)
    log_port("function_ubee_r", "id char", port, c);

 return (uint16_t)(c);
}

//==============================================================================
// Status functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_status (int addr)
{
 int i;
 int x;
 char vers[6];
 int res; 
 
 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);
 
 switch (leu16_to_host(f->cmdres.cmd))
    {
     case 0x00 : // Get version information
                 i = 0;
                 
                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 f->version.ver1 = host_to_leu16(res);
                 i++;

                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 f->version.ver2 = host_to_leu16(res);
                 i++;

                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 f->version.ver3 = host_to_leu16(res);
                 i++;

                 break;

     case 0x01 : // Get host system information
#ifdef MINGW
                 f->cmdres.res = host_to_leu16(0);	// win32
#else
                 f->cmdres.res = host_to_leu16(1);	// unices
#endif                     
                 break;
     case 0x02 : // Get host integer size
                 f->cmdres.res = host_to_leu16(sizeof(i));
                 break;
     case 0x03 : // Get host endian format
                 f->cmdres.res = host_to_leu16(IsBigEndian());
                 break;

     case 0x10 : // Get model number emulated (subject to change)
                 f->cmdres.res = host_to_leu16(emu.model);
                 break;
     case 0x11 : // Get alpha+ model emulation status
                 f->cmdres.res = host_to_leu16(modelx.alphap);
                 break;
     case 0x12 : // Get ROM model emulation status
                 f->cmdres.res = host_to_leu16(modelx.rom);
                 break;                 
     case 0x13 : // Get model boot address status
                 f->cmdres.res = host_to_leu16(modelx.bootaddr);
                 break;
     case 0x14 : // Get RAM emulation size
                 f->cmdres.res = host_to_leu16(modelx.ram);
                 break;                 
     case 0x15 : // Get colour emulation status
                 f->cmdres.res = host_to_leu16(modelx.colour);
                 break;
     case 0x16 : // Get hardware flash emulation status
                 f->cmdres.res = host_to_leu16(modelx.hwflash);
                 break;
     case 0x17 : // Get light pen keys emulation status
                 f->cmdres.res = host_to_leu16(modelx.lpen);
                 break;
     case 0x18 : // Get speed switch emulation status
                 f->cmdres.res = host_to_leu16(modelx.speed);
                 break;
     case 0x19 : // Get speed piob7 emulation status
                 f->cmdres.res = host_to_leu16(modelx.piob7);
                 break;
     case 0x1A : // Get RTC emulation status
                 f->cmdres.res = host_to_leu16(modelx.rtc);
                 break;
       default : break;
    }
}    

//==============================================================================
// Control functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_control (int addr)
{
 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);
 
 if (leu16_to_host(f->cmdres.id) == 0xAA55)
    {
     f->file.id = host_to_leu16(0);  // clear it for next time (must be set each time)
     switch (leu16_to_host(f->cmdres.cmd))
        {
         case 0x00 : emu.reset = 2;	// action 2 is reset, no confirmation
                     break;
         case 0x01 : emu.done = 1;
                     break;
         case 0x02 : crtc.video = 1;	// turn the video on
                     break;
         case 0x03 : sound.mute = 0;	// turn the sound on
                     break;
         case 0x04 : set_clock_speed(emu.cpuclock_def, emu.z80_divider, 0);
                     break;
           default : break;
        }
    }
}    

//==============================================================================
// Diagnostics functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_diagnostics (int addr)
{
 int addr1;
 int lines;
 int htype; 

 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);
 
 if (leu16_to_host(f->cmdres.id) == 0xAA55)
    {
     f->file.id = host_to_leu16(0);  // clear it for next time (must be set each time)
     switch (leu16_to_host(f->cmdres.cmd))
        {
         case 0x00 : addr1 = leu16_to_host(f->dump.addr);
                     lines = leu16_to_host(f->dump.lines);
                     htype = leu16_to_host(f->dump.htype);
                     z80debug_dump(addr1, lines, htype);
                     break;
           default : break;
        }
    }
}

//==============================================================================
// Keyboard functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_keyboard (int addr)
{
// ub_func_u *f = (ub_func_u*)(z80mem_f + addr);
}    

//==============================================================================
// stdio functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_stdio (int addr)
{
 int addr1;

 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);

 switch (leu16_to_host(f->cmdres.cmd))
    {
     case 0x00 : // getchar function
                 f->getchar.res = host_to_le16(getchar());
                 break;
     case 0x01 : // putchar function
                 f->putchar.res = host_to_le16(putchar(le16_to_host(f->putchar.val)));
                 break;
     case 0x02 : // puts function
                 addr1 = leu16_to_host(f->puts.addr);
                 z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                 f->puts.res = host_to_le16(puts((char*)(z80mem_x1+addr1)));
                 break;
     case 0x03 : // printf function (no variables)
                 addr1 = leu16_to_host(f->puts.addr);
                 z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                 xprintf("%s", (char*)(z80mem_x1+addr1));
                 break;

     case 0x80 : // write 16 bit value as signed decimal
                 xprintf("%d", le16_to_host(f->putchar.val));
                 break;
     case 0x81 : // write 16 bit value as unsigned decimal
                 xprintf("%ud", leu16_to_host(f->putchar.val));
                 break;
     case 0x82 : // write 16 bit value as 2 hex characters
                 xprintf("%02x", leu16_to_host(f->putchar.val));
                 break;
     case 0x83 : // write 16 bit value as 4 hex characters
                 xprintf("%04x", leu16_to_host(f->putchar.val));
                 break;

       default : break;
    }
}    

//==============================================================================
// File functions.
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_files (int addr)
{
 int amount_t;
 int amount_l;
 int amount_h;
 int inrange;
 
 int addr1;
 int addr2;
 int addr3;
 int z80_addr;

 char *ptrres;

 char buffer[0x10000];

 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);

 if (leu16_to_host(f->file.id) == 0xAA55)
    {
     f->file.id = host_to_leu16(0);  // clear it for next time (must be set each time)
     switch (leu16_to_host(f->file.cmd))
        {
         case 0x00 : // fopen
                     addr1 = leu16_to_host(f->file.addr1);
                     addr2 = leu16_to_host(f->file.addr2);
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     z80mem_x2 = get_z80mem_ptr_and_addr(&addr2);
                     f->file.fp.p = fopen((char*)(z80mem_x1+addr1), (char*)(z80mem_x2+addr2));
                     f->file.res = host_to_le16(f->file.fp.p == NULL);
                     break;
         case 0x01 : // fclose
                     f->file.res = host_to_le16(fclose(f->file.fp.p));
                     break;
         case 0x02 : // fflush
                     f->file.res = host_to_le16(fflush(f->file.fp.p));
                     break;
         case 0x03 : // feof
                     f->file.res = host_to_le16(feof(f->file.fp.p));
                     break;
         case 0x04 : // ferror
                     f->file.res = host_to_le16(ferror(f->file.fp.p));
                     break;
         case 0x05 : // fgetc
                     f->file.res = host_to_le16(fgetc(f->file.fp.p));
                     break;
         case 0x06 : // fgetpos (RESERVED)
                     break;
         case 0x07 : // fgets
                     addr3 = leu16_to_host(f->file.addr3);
                     z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);
                     ptrres = fgets((char*)(z80mem_x3+addr3), leu16_to_host(f->file.num), f->file.fp.p);
                     f->file.res = host_to_leu16(ptrres != NULL);
                     break;
         case 0x08 : // fputc
                     f->file.res = host_to_le16(fputc(leu16_to_host(f->file.val1), f->file.fp.p));
                     break;
         case 0x09 : // fputs
                     addr3 = leu16_to_host(f->file.addr3);
                     z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);
                     f->file.res = host_to_le16(fputs((char*)(z80mem_x3+addr3), f->file.fp.p));
                     break;                     
         case 0x0A : // fread
                     if ((leu16_to_host(f->file.size) * leu16_to_host(f->file.num)) <= sizeof(buffer))
                        {
                         amount_t = fread(buffer, leu16_to_host(f->file.size), leu16_to_host(f->file.num), f->file.fp.p);
                         f->file.res = host_to_le16(amount_t);  // save the result of the read (bytes read)
                         addr3 = leu16_to_host(f->file.addr3);
                         z80_addr = addr3;
                         if (z80_addr < 0x8000)  // if destination is to a low bank
                            {
                             z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current lower bank pointer
                             amount_l = 0x8000 - addr3;  // amount to be moved to this bank
                             if (amount_t < amount_l)    // if the amount read is less than max space then use that instead
                                amount_l = amount_t;
                             amount_h = amount_t - amount_l;
                             
                             inrange = (amount_h <= 0x8000);
                             if (inrange)
                                {
                                 memcpy(z80mem_x3+addr3, buffer, amount_l);
                                 addr3 = 0x8000;
                                 z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current upper bank pointer
                                 memcpy(z80mem_x3, buffer+amount_l, amount_h);
                                } 
                            }
                         else   
                            {
                             amount_h = amount_t;
                             amount_l = 0;  // needed for error checking!
                             inrange = ((z80_addr + amount_h) <= 0x10000);
                             if (inrange)
                                {
                                 z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current upper bank pointer
                                 memcpy(z80mem_x3+addr3, buffer, amount_h);
                                } 
                            }
                         if (modio.func)
                            {
                             log_data_5("function_files", "function", "Z80 addr", "read", "amount(LB)", "amount(HB)",
                             leu16_to_host(f->file.cmd), z80_addr, amount_t, amount_l, amount_h);
                            }
                         if (! inrange)
                            {
                             log_data_3("function_files (data exceeds 32K limit)", "function", "amount(LB)", "amount(HB)",
                             leu16_to_host(f->file.cmd), amount_l, amount_h);
                             f->file.res = host_to_le16(0);
                            }
                        }
                     else
                        {
                         xprintf("function_files: function=0x%04x: internal buffer exceeded\n", leu16_to_host(f->file.cmd));
                         f->file.res = host_to_le16(0);
                        } 
                     break;
         case 0x0B : // freopen (RESERVED)
                     break;
         case 0x0C : // fseek (RESERVED)
                     break;
         case 0x0D : // fsetpos (RESERVED)
                     break;
         case 0x0E : // ftell (RESERVED)
                     break;
         case 0x0F : // fwrite
                     amount_t = leu16_to_host(f->file.size) * leu16_to_host(f->file.num);
                     if (amount_t <= sizeof(buffer))
                        {
                         addr3 = leu16_to_host(f->file.addr3);
                         z80_addr = addr3;
                         if (z80_addr < 0x8000)  // if source is from a low bank
                            {
                             z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current lower bank pointer
                             if ((addr3 + amount_t) > 0x8000)  // amount to be moved from this bank
                                amount_l = 0x8000 - addr3;
                             else
                                amount_l = amount_t;   
                             amount_h = amount_t - amount_l;
                             memcpy(buffer, z80mem_x3+addr3, amount_l);
                             
                             inrange = (amount_h <= 0x8000);
                             if (inrange)
                                {
                                 addr3 = 0x8000;
                                 z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current upper bank pointer
                                 memcpy(buffer+amount_l, z80mem_x3, amount_h);
                                } 
                            }
                         else   
                            {
                             amount_h = amount_t;
                             amount_l = 0;  // needed for error checking!
                             inrange = ((z80_addr + amount_h) <= 0x10000);
                             if (inrange)
                                {
                                 z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);  // get the current upper bank pointer
                                 memcpy(buffer, z80mem_x3+addr3, amount_h);
                                } 
                            }
                         if (modio.func)
                            {
                             log_data_5("function_files", "function", "Z80 addr", "write", "amount(LB)", "amount(HB)",
                             leu16_to_host(f->file.cmd), z80_addr, amount_t, amount_l, amount_h);
                            }
                         if (! inrange)
                            {
                             log_data_3("function_files (data exceeds 32K limit)", "function", "amount(LB)", "amount(HB)",
                             leu16_to_host(f->file.cmd), amount_l, amount_h);
                             f->file.res = host_to_le16(0);
                            }
                         else
                            f->file.res = host_to_le16(fwrite(buffer, leu16_to_host(f->file.size), leu16_to_host(f->file.num), f->file.fp.p));
                        }
                     else
                        {
                         xprintf("function_files: function=0x%04x: internal buffer exceeded\n", leu16_to_host(f->file.cmd));
                         f->file.res = host_to_le16(0);
                        } 
                     break;
         case 0x10 : // clearerr
                     clearerr(f->file.fp.p);
                     break;
         case 0x11 : // rewind
                     rewind(f->file.fp.p);
                     break;
         case 0x12 : // remove
                     addr1 = leu16_to_host(f->file.addr1);
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     f->file.res = host_to_le16(remove((char*)(z80mem_x1+addr1)));
                     break;
         case 0x13 : // rename
                     addr1 = leu16_to_host(f->file.addr1);
                     addr2 = leu16_to_host(f->file.addr2);
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     z80mem_x2 = get_z80mem_ptr_and_addr(&addr2);
                     f->file.res = host_to_le16(rename((char*)(z80mem_x1+addr1),
                     (char*)(z80mem_x2+addr2)));
                     break;
         case 0x14 : // tmpfile
                     f->file.fp.p = tmpfile();
                     break;
         case 0x15 : // tmpnam (RESERVED)
                     break;
         case 0x16 : // ugetc (RESERVED)
                     break;
         case 0x17 : // eofval constant
                     f->file.res = host_to_le16(EOF);
                     break;
           default : break;
        }
    }
}       

//==============================================================================
// Directory functions.
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_directory (int addr)
{
 char const illegal_cpm_char[] = "<>,;:=?*[]\%|()/\\";
 struct dirent *de;
 struct stat st;

 char filename_dir[1024];
 char filename_search[1024]; 

 char filepath[256]; 
 char filename[256];

 int l;
 int i;
 int addr1;
 int addr2;
 int addr3;
 int addr4;

 int res;
 int dot;
 int match;
 int dir_type;

 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);
   
 if (leu16_to_host(f->file.id) == 0xAA55)
    {
     f->file.id = host_to_leu16(0);  // clear it for next time (must be set each time)
     switch (leu16_to_host(f->file.cmd))
        {
         case 0x00 : // open a directory for reading
                     addr1 = leu16_to_host(f->file.addr1); // pass directory path/name
                     addr2 = leu16_to_host(f->file.addr2); // return file name or wildcards
                     addr3 = leu16_to_host(f->file.addr3); // return modified file path here
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     z80mem_x2 = get_z80mem_ptr_and_addr(&addr2);
                     z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);
                     z80mem_x2[addr2] = 0;
                     f->file.val1 = host_to_leu16(0);
                     res = 1;

                     strcpy(filepath, (char*)(z80mem_x1+addr1));
                     l = strlen(filepath);

                     if (! l)                                  // add '.' if an empty path
                        strcat(filepath, ".");
                     else   
                        if (strcmp(filepath, "."SLASHCHAR_STR) == 0)  // if './' or '.\'
                           filepath[l-1] = 0;                  // remove the slash character
#ifdef MINGW
                        else
                           if (filepath[l-1] == ':')           // if last char is ':'
                              strcat(filepath, SLASHCHAR_STR); // add a '\'
#endif                              

                     if (stat(filepath, &st) == 0)
                        {
                         res = ((st.st_mode & S_IFDIR) == 0);
                         if (! res)
                            {
                             z80mem_x2[addr2] = '*';
                             z80mem_x2[addr2+1] = 0;
                            }
                        } 
                     else
                        res = 1;   

                     if (res)   
                        {
                         l = strlen(filepath);
#ifdef MINGW                         
                         while ((l) && (filepath[l-1] != SLASHCHAR) && (filepath[l-1] != ':'))
#else
                         while ((l) && (filepath[l-1] != SLASHCHAR))
#endif
                            l--;

                         strcpy((char*)(z80mem_x2+addr2), filepath+l); // copy the name part
                         if ((l > 1) && (filepath[l-1] == SLASHCHAR))  // if '/' or '\'
                            l--;
                         filepath[l] = 0;                         // remove the name part

                         if (! strlen((char*)z80mem_x2+addr2))       // if file name is blank
                            {
                             z80mem_x2[addr2] = '*';
                             z80mem_x2[addr2+1] = 0;
                            }                         

                         l = strlen(filepath);
                         if (! l)                                 // add '.' if empty path
                            strcat(filepath, ".");

                         if (stat(filepath, &st) == 0)            // try again without name
                            res = ((st.st_mode & S_IFDIR) == 0);
                         else
                            res = 1;       
                        }

                     if (! res)
                        {
                         f->file.fp.d = opendir(filepath);

                         l = strlen(filepath);
                         if ((l) && (filepath[l-1] != SLASHCHAR))  // add '/' or '\' if no end slash
                            strcat(filepath, SLASHCHAR_STR);

                         res = (f->file.fp.d == NULL);
                        }
                     
                     f->file.val1 = host_to_le16(strlen((char*)z80mem_x2+addr2));
                     strcpy((char*)z80mem_x3+addr3, filepath);  // return the file path
                     f->file.val2 = host_to_le16(strlen(filepath));
                     f->file.res = host_to_le16(res);
                     
                     break;    
         case 0x01 : // close directory
                     f->file.res = host_to_le16(closedir(f->file.fp.d));
                     break;
         case 0x02 : // return the next directory entry (with wildcards)
                     addr1 = leu16_to_host(f->file.addr1); // return name match
                     addr2 = leu16_to_host(f->file.addr2); // pass file name or wildcards
                     addr3 = leu16_to_host(f->file.addr3); // pass file path
                     addr4 = leu16_to_host(f->file.addr4); // return full path and file name match
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     z80mem_x2 = get_z80mem_ptr_and_addr(&addr2);
                     z80mem_x3 = get_z80mem_ptr_and_addr(&addr3);
                     z80mem_x4 = get_z80mem_ptr_and_addr(&addr4);
                     strcpy(filepath, (char*)z80mem_x3+addr3);
                     match = 0;
                     
                     strncpy(filename_search, (char*)z80mem_x2+addr2, sizeof(filename_search));
                     filename_search[sizeof(filename_search)-1] = 0;
#ifdef MINGW
                     toupper_string(filename_search, filename_search);
#endif
                     do
                        {
                         de = readdir(f->file.fp.d);
                         if (de)
                            {
                             strncpy(filename_dir, de->d_name, sizeof(filename_dir));
                             filename_dir[sizeof(filename_dir)-1] = 0;
#ifdef MINGW
                             toupper_string(filename_dir, filename_dir);
#endif
                             match = wildcardfit(filename_search, filename_dir);

                             if (match)
                                {
                                 strcpy((char *)z80mem_x1+addr1, de->d_name);
                                 snprintf(filename, sizeof(filename), "%s%s", filepath, de->d_name);
                                 if (stat(filename, &st) == 0)
                                    {
                                     dir_type = 0;
#ifdef MINGW
#else
                                     dir_type |= (S_ISLNK(st.st_mode) != 0);
#endif                                     
                                     dir_type |= ((S_ISREG(st.st_mode) != 0) << 1);
                                     dir_type |= ((S_ISDIR(st.st_mode) != 0) << 2);
                                     dir_type |= ((S_ISCHR(st.st_mode) != 0) << 3);
                                     dir_type |= ((S_ISBLK(st.st_mode) != 0) << 4);
                                     dir_type |= ((S_ISFIFO(st.st_mode) != 0) << 5);
#ifdef MINGW
#else
                                     dir_type |= ((S_ISSOCK(st.st_mode) != 0)  << 6);
#endif                                        
                                    }
                                 else
                                    dir_type = 0;
                                 strcpy((char *)z80mem_x4+addr4, filename);
                                 f->file.val2 = host_to_le16(strlen(filename));                                 
                                 f->file.res = host_to_le16(dir_type);
                                 f->file.val1 = host_to_le16(strlen((char*)z80mem_x1+addr1));
                                }
                            }    
                        }
                     while ((de) && (! match));
                     
                     if (! match)
                        {
                         z80mem_x1[addr1] = 0;
                         z80mem_x4[addr4] = 0;                         
                         f->file.res = host_to_leu16(0);
                         f->file.val1 = host_to_leu16(0);
                         f->file.val2 = host_to_leu16(0);                         
                        }
                     break;
         case 0x03 : // wildcards test
                     z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                     z80mem_x2 = get_z80mem_ptr_and_addr(&addr2);
                     addr1 = leu16_to_host(f->file.addr1); // pass file name to be tested
                     addr2 = leu16_to_host(f->file.addr2); // pass wildcards
                     f->file.res = host_to_le16(wildcardfit((char*)(z80mem_x2+addr2), (char*)(z80mem_x1+addr1)));
                     break;
         case 0x04 : // check if filename is a legal CP/M 8.3 format
                     addr1 = leu16_to_host(f->file.addr1); // pass file name to be tested
                     res = 0;
                     dot = 0;
                     i = 0;
                     l = strlen((char*)z80mem_x1+addr1);
                     res = (l > 12);
                     if (! res)
                        {
                         while ((i < l) && (! res))
                            {
                             if (z80mem_x1[addr1+i] == '.')
                                {
                                 dot++;
                                 res = (dot > 1) * 2;
                                 if ((! res) && (l > 3))
                                    res = (i < (l-4)) * 3;
                                } 
                             else
                                res = (strchr(illegal_cpm_char, z80mem_x1[addr1+i]) != NULL) * 4;
                             i++;
                            }
                        }
                     if ((! res) && (! dot) && (l > 8))
                        res = 5;
                     f->file.res = host_to_le16(res);                        
           default : break;
        }
    }
}       

//==============================================================================
// Joystick functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_joystick (int addr)
{
 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);

 switch (leu16_to_host(f->cmdres.cmd))
    {
     case 0x00 : // enable/disable Microbee joystick
                 joystick.mbee = le16_to_host(f->putchar.val);
                 break;                 
     case 0x01 : // enable/disable joystick keys mapping
                 joystick.kbd = le16_to_host(f->putchar.val);
                 break;
     case 0x02 : // select a joystick keys mapping set
                 joystick_kbjoy_select(le16_to_host(f->putchar.val), "");
                 break;
     case 0x03 : // upload buttons for the Microbee joystick (RESERVED)
                 break;
     case 0x04 : // upload buttons for a joystick key mapping set (RESERVED)
                 break;
       default : break;
    }
}    

//==============================================================================
// Application functions
//
//   pass: int addr             address of Z80 structure (relative to bank)
// return: void
//==============================================================================
static void function_application (int addr)
{
 int addr1;

 ub_func_u *f = (ub_func_u*)(z80mem_f + addr);

 switch (leu16_to_host(f->cmdres.cmd))
    {
     case 0x00 : // return a --file-list option parameter string
                 addr1 = leu16_to_host(f->puts.addr);
                 z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                 strcpy((char*)z80mem_x1+addr1, func.file_list[func.file_list_pos]);
                 f->puts.res = host_to_le16(strlen((char*)z80mem_x1+addr1));
                 break;                 
     case 0x01 : // return the --file-run option parameter string
                 addr1 = leu16_to_host(f->puts.addr);
                 z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                 strcpy((char*)z80mem_x1+addr1, func.file_run);
                 f->puts.res = host_to_le16(strlen((char*)z80mem_x1+addr1));
                 break;
     case 0x02 : // return the --file-app option parameter string
                 addr1 = leu16_to_host(f->puts.addr);
                 z80mem_x1 = get_z80mem_ptr_and_addr(&addr1);
                 strcpy((char*)z80mem_x1+addr1, func.file_app);
                 f->puts.res = host_to_le16(strlen((char*)z80mem_x1+addr1));
                 break;
     case 0x03 : // return the --file-load option Z80 address
                 f->puts.res = host_to_le16(func.file_load);
                 break;
     case 0x04 : // return the --file-exec option Z80 address
                 f->puts.res = host_to_le16(func.file_exec);
                 break;
     case 0x05 : // return the --file-exit option status
                 f->puts.res = host_to_le16(func.file_exit);
                 break;
     case 0x06 : // return the --file-list string count
                 f->puts.res = host_to_le16(func.file_list_count);
                 break;                 

     case 0x10 : // set the current --file-list position
                 if (le16_to_host(f->putchar.val) < (func.file_list_count))
                    func.file_list_pos = le16_to_host(f->putchar.val);
                 else
                    func.file_list_pos = 0;
                 break;                 
       default : break;
    }
}    

//==============================================================================
// Basic functions
//
// Any functions numbers that are not called in function_ubee_w() end up here.
// These functions are acted on directly and have no parameters and return no
// results.
//
// New functions here should only be added when a quick and easy method is
// required to perform a task.  The advantage is that no table structure is
// needed but the number of possible functions are limited.
//
//   pass: int cmd
// return: void
//==============================================================================
static void function_basic (int cmd)
{
 switch (cmd)
    {
     case 0xe0 : // turn on video, sound and revert to default clock rate
                 crtc.video = 1;	// turn the video on
                 sound.mute = 0;	// turn the sound on
                 set_clock_speed(emu.cpuclock_def, emu.z80_divider, 0);
                 break;     
       default : break;
    }
}    

//==============================================================================
// uBee512 write - Special uBee512 Port function
//
// All commands pass a Z80 address to a structure in memory.  The structure
// contents is dependent on the command but all share the same overhead.  Data
// passed and returned in the structure is always in little endian format.
//
// IMPORTANT: This function does not currently address banking issues so it
// is important to realise that data may get lost if ANY bank swapping occurs
// after this function call.  To avoid problems do not switch banks or call
// any OS calls before the returned data is used.  Essentially the main area
// of concern is the 32K DRAM bank at 0x8000 that may also be at 0x0000.
// Normally writes to these banks will be copied from one to the other, but
// this does not occur here as we are acting on Z80 memory outside of the
// MZ80 context.  Also do not have the structure in video memory for similar
// reasons,  unlikely you would want to do that any way.
//
// It is important to verify that the application is running under the emulator
// and not a real Microbee before calling this function.  Use the uBee512 read
// port function to check first.
//
// The calling sequence is:
// port[n] = command number (0..255)
// port[n] = LSB of structure address (0 if not using a structure) 
// port[n] = MSB of structure address (0 if not using a structure)
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void function_ubee_w (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 if (modio.func)
    log_port("function_ubee_w", "data", port, data);    

 if (ubee_req)		// if we need to get the structure address
    {
     ubee_req--;
     if (ubee_req == 1)
        ubee_addr = data;	// LSB of the address
     else   
        if (ubee_req == 0)
           {
            ubee_addr = ubee_addr | (data << 8);
            z80mem_f = get_z80mem_ptr_and_addr(&ubee_addr);

            switch (ubee_command)
               {
                case 0x00 : // status functions
                            function_status(ubee_addr);
                            break;
                case 0x10 : // control functions
                            function_control(ubee_addr);
                            break;
                case 0x20 : // diagnostics functions
                            function_diagnostics(ubee_addr);
                            break;
                case 0x30 : // keyboard functions
                            function_keyboard(ubee_addr);
                            break;                            
                case 0x40 : // stdio functions
                            function_stdio(ubee_addr);
                            break;                            
                case 0x50 : // file functions
                            function_files(ubee_addr);
                            break;
                case 0x60 : // directory functions
                            function_directory(ubee_addr);
                            break;
                case 0x70 : // joystick functions
                            function_joystick(ubee_addr);
                            break;
                case 0x80 : // application functions
                            function_application(ubee_addr);
                            break;
                  default : function_basic(ubee_command);
                            break;          
               }
           }
    }
 else
    {
     ubee_command = data;
     ubee_req = 2;
    }
}
