//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                              Joystick module                               *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// This module is used to provide joystick support
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.1.0 - 22 June 2009, S.J.Kay
// - Made improvements to the logging process.
//
// v4.0.0 - 13 May 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
//
// v3.1.0 - 7 November 2008, S.J.Kay
// - Changed functions function_*() to keyb_*() as code was moved.
// - Changed all printf() calls to use a local xprintf() function.
//
// v3.0.0 - 24 September 2008, S.J.Kay
// - Added C_WINI and C_WIND commands for increasing/decreasing window size.
// - Added C_GLFILT command for toggling of OpenGL filter mode, and C_VSIZE1
//   command for video resizing.
// - Added C_MWHEEL command to select mouse wheel association.
//
// v2.8.0 - 3 September 2008, S.J.Kay
// - JS key commands can now be held down and the commands repeated.
// - Added mapping of joystick buttons to emulator (EMUKEY style) commands and
//   joystick_kbjoy_listcommands() function to list the command names. 
//
// v2.7.0 - 4 June 2008, S.J.Kay
// - Made Hat motion reportable if +joystick used in --modio option.
// - Fixed 12 button limitation, now is 128 buttons as originally intended.
// - Added structure emu_t and moved variables into it.
//
// v2.6.0 - 10 May 2008, S.J.Kay
// - Created a new file and implement the joystick emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <SDL.h>

#include "joystick.h"
#include "function.h"
#include "keyb.h"
#include "keystd.h"
#include "ubee512.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
joystick_t joystick =
{
 .joy = NULL,
 .used = -1,
 .mbee = 1,
 .axis_buttons = 0,
 .axis_level = 3200
};

static int cmd_scan;

static int axis_button_left;
static int axis_button_right;
static int axis_button_up;
static int axis_button_down;

extern emu_t emu;
extern modio_t modio;

//==============================================================================
// structures and variables related to Microbee Joystick emulation (PIO A)
//==============================================================================
static uint8_t mb_button_data[128] =
{
 JOY_MB_UP,		// 0
 JOY_MB_RIGHT,		// 1
 JOY_MB_DOWN,		// 2
 JOY_MB_LEFT,		// 3
 JOY_MB_FIRE,		// 4
 JOY_MB_PLAY1,		// 5
 JOY_MB_PLAY2,		// 6
 JOY_MB_SPARE		// 7
};

//==============================================================================
// structures and variables related to Joystick to 6545 key and emulator
// commands mappings.
//==============================================================================
static int kb_joy_init;

static int kb_scan;

static uint8_t kb_button_data[JOY_KB_ENTRIES][128];

static uint8_t *kb_button_data_use = *kb_button_data;

static uint8_t kb_button_data_buf[128];

// Search table of Microbee keys and corresponding matrix codes (6545 CRTC)
static char *key_to_6545[] =
{
 "@",     "A",     "B",     "C",     "D",     "E",     "F",     "G",
 "H",     "I",     "J",     "K",     "L",     "M",     "N",     "O",
 "P",     "Q",     "R",     "S",     "T",     "U",     "V",     "W",
 "X",     "Y",     "Z",     "[",     "\\",    "]",     "^",     "DEL",
 "0",     "1",     "2",     "3",     "4",     "5",     "6",     "7",
 "8",     "9",     ":",     ";",     ",",     "-",     ".",     "/",
 "ESC",   "BS",    "TAB",   "LF",    "CR",    "LOCK",  "BRK",   "SP",
 "UP",    "CTRL",  "DOWN",  "LEFT",  "60",    "61",    "RIGHT", "SHIFT",
 ""
};

// Search table of emulator commands.  This must match the EMU_CMD_NAME
// number ordering.  If any commands are not supported then padding is still
// required.
static char *emu_commands[] =
{
 "C_FSTOG",
 "C_TAPER",
 "C_DMP",
 "C_DMPN1",
 "C_DMPN2",
 "C_DMPB1",
 "C_DMPB2",
 "C_DMPREP",
 "C_DMPREG",
 "C_DBOFF",
 "C_DBON",
 "C_DBTRA",
 "C_DBST1",
 "C_DBST10",
 "C_DBST20",
 "C_JS",
 "C_MUTE",
 "C_PAUSE",
 "C_VOLU",
 "C_VOLD",
 "C_WINI",
 "C_WIND",
 "C_VSIZE1",
 "C_GLFILT",
 "C_MWHEEL",
 ""
};

//==============================================================================
// Joystick Initialise. 
//
//   pass: void
// return: int				0 if no error, else -1
//==============================================================================
int joystick_init (void)
{
 joystick.data = B8(11111111);

 if (joystick.used >= 0)
    {
     SDL_JoystickEventState(SDL_ENABLE);

     joystick.joy = SDL_JoystickOpen(joystick.used);

     if (joystick.joy)
        {
         if (modio.joystick)
            {
             xprintf("joystick_init:\n");
             xprintf("Joystick: %d\n", joystick.used);
             xprintf("Name: %s\n", SDL_JoystickName(joystick.used));
             xprintf("Number of Axes: %d\n", SDL_JoystickNumAxes(joystick.joy));
             xprintf("Number of Buttons: %d\n", SDL_JoystickNumButtons(joystick.joy));
             xprintf("Number of Balls: %d\n", SDL_JoystickNumBalls(joystick.joy));

             if (modio.level)
                {
                 fprintf(modio.log, "joystick_init:\n");
                 fprintf(modio.log, "Joystick: %d\n", joystick.used);
                 fprintf(modio.log, "Name: %s\n", SDL_JoystickName(joystick.used));
                 fprintf(modio.log, "Number of Axes: %d\n", SDL_JoystickNumAxes(joystick.joy));
                 fprintf(modio.log, "Number of Buttons: %d\n", SDL_JoystickNumButtons(joystick.joy));
                 fprintf(modio.log, "Number of Balls: %d\n", SDL_JoystickNumBalls(joystick.joy));
                }
            }
        }
     else
        xprintf("joystick_init: Failed to open joystick device: %d\n", joystick.used);
    }

 return 0;
}

//==============================================================================
// Joystick de-initialise.
//
// For joystick out the WAV file header is filled in and the file closed.
// For joystick in the file is closed.
//
//   pass: void
// return: int				0
//==============================================================================
int joystick_deinit (void)
{
 if (joystick.joy)
    SDL_JoystickClose(joystick.joy);

 return 0;
}

//==============================================================================
// Joystick reset.
//
//   pass: void
// return: int				0
//==============================================================================
int joystick_reset (void)
{
 if (joystick.joy)
    {
    } 

 return 0;
}

//==============================================================================
// Joystick button down
//
//   pass: int button
// return: void
//==============================================================================
static void joystick_buttondown (int button)
{
 int scan;
 
 if (button < 128)
    {
     joystick.data = joystick.data ^ mb_button_data[button];

     if (joystick.kbd)
        {
         scan = kb_button_data_use[button];
         if (scan)
            {
             if (scan < 65)
                keystd_scan_set(scan-1);
             else
                if ((scan-65) != EMU_CMD_JOYSTICK)
                   {
                    cmd_scan = scan;
                    keyb_repeat_start();
                    keyb_emu_command(scan-65, 0);
                   } 
            }   
        }    
    } 
}

//==============================================================================
// Joystick button down event
//
//   pass: void
// return: void
//==============================================================================
void joystick_buttondown_event (void)
{
 joystick_buttondown(emu.event.jbutton.button);
 
 if (modio.joystick)
    log_data("joystick_buttondown_event", "button", emu.event.jbutton.button);
} 

//==============================================================================
// Joystick button up
//
//   pass: int button
// return: void
//==============================================================================
static void joystick_buttonup (int button)
{
 int scan;

 if (button < 128)
    {
     joystick.data = joystick.data | mb_button_data[button];

     if (joystick.kbd)
        {
         scan = kb_button_data_use[button];
         if ((scan) && (scan < 65))
            keystd_scan_clear(scan-1);
         else
            {
             if (cmd_scan == scan)
                keyb_repeat_stop();
            }   
        }
    }    
}

//==============================================================================
// Joystick button up event
//
//   pass: void
// return: void
//==============================================================================
void joystick_buttonup_event (void)
{
 joystick_buttonup(emu.event.jbutton.button);

 if (modio.joystick)
    log_data("joystick_buttonup_event", "button", emu.event.jbutton.button);
} 

//==============================================================================
// Joystick hat motion event
//
//   pass: void
// return: void
//==============================================================================
void joystick_hatmotion_event (void)
{
 if (modio.joystick)
    {
     if (emu.event.jhat.value & SDL_HAT_UP)
        log_mesg("joystick_hatmotion_event: Hat up");
     if (emu.event.jhat.value & SDL_HAT_DOWN)
        log_mesg("joystick_hatmotion_event: Hat down");
     if (emu.event.jhat.value & SDL_HAT_LEFT)
        log_mesg("joystick_hatmotion_event: Hat left");
     if (emu.event.jhat.value & SDL_HAT_RIGHT)
        log_mesg("joystick_hatmotion_event: Hat right");
     if (emu.event.jhat.value & SDL_HAT_CENTERED)
        log_mesg("joystick_hatmotion_event: Hat centered");
    }
} 

//==============================================================================
// Joystick Axis motion event
//
// The joystick may need to be calibrated to work correctly.
//
//   pass: void
// return: void
//==============================================================================
void joystick_axismotion_event (void)
{
 if (modio.joystick)
    {
     if (emu.event.jaxis.axis == 0) 
        log_data_2("joystick_axismotion_event", "X axis", "X axis value",
        emu.event.jaxis.axis, emu.event.jaxis.value);

     if (emu.event.jaxis.axis == 1) 
        log_data_2("joystick_axismotion_event", "Y axis", "Y axis value",
        emu.event.jaxis.axis, emu.event.jaxis.value);
    }

// X axis movement
 if (emu.event.jaxis.axis == 0)
    {
     if ((emu.event.jaxis.value >= -joystick.axis_level) && (axis_button_left))
        {
         axis_button_left = 0;
         joystick_buttonup(joystick.axis_buttons + 3);
        } 
     if ((emu.event.jaxis.value <= joystick.axis_level) && (axis_button_right))
        {
         axis_button_right = 0;
         joystick_buttonup(joystick.axis_buttons + 1);
        }

     if ((emu.event.jaxis.value < -joystick.axis_level) && (! axis_button_left))
        {
         if (axis_button_right)
            {
             axis_button_right = 0;
             joystick_buttonup(joystick.axis_buttons + 1);             
            }
         axis_button_left = 1;
         joystick_buttondown(joystick.axis_buttons + 3);
         return;
        }

    if ((emu.event.jaxis.value > joystick.axis_level) && (! axis_button_right))
       {
        if (axis_button_left)
           {
            axis_button_left = 0;
            joystick_buttonup(joystick.axis_buttons + 3);             
           }
        axis_button_right = 1;
        joystick_buttondown(joystick.axis_buttons + 1);
        return;
       } 
    }

// Y axis movement
 if (emu.event.jaxis.axis == 1)
    {
     if ((emu.event.jaxis.value >= -joystick.axis_level) && (axis_button_up))
        {
         axis_button_up = 0;
         joystick_buttonup(joystick.axis_buttons + 0);
        }
     if ((emu.event.jaxis.value <= joystick.axis_level) && (axis_button_down))
        {
         axis_button_down = 0;
         joystick_buttonup(joystick.axis_buttons + 2);
        }

     if ((emu.event.jaxis.value < -joystick.axis_level) && (! axis_button_up))
        {
         if (axis_button_down)
            {
             axis_button_down = 0;
             joystick_buttonup(joystick.axis_buttons + 2);             
            }
         axis_button_up = 1;
         joystick_buttondown(joystick.axis_buttons + 0);
         return;
        }

     if ((emu.event.jaxis.value > joystick.axis_level) && (! axis_button_down))
        {
         if (axis_button_up)
            {
             axis_button_up = 0;
             joystick_buttonup(joystick.axis_buttons + 0);             
            }
         axis_button_down = 1;
         joystick_buttondown(joystick.axis_buttons + 2);
         return;
        } 
    }
} 

//==============================================================================
// Clear all the Microbee joystick button actions.
//
//   pass: void
// return: void
//==============================================================================
void joystick_mbjoy_clear (void)
{
 memset(mb_button_data, 0, sizeof(mb_button_data));
} 

//==============================================================================
// Set the Microbee joystick button to be associated with a joystick action.
//
//   pass: int action
//         int button
// return: int				0 if no error, else -1
//==============================================================================
int joystick_mbjoy_set_action (int action, int button)
{
 if ((action < 0) || (action > 255) || (button < 0) || (button >= sizeof(mb_button_data)))
    return -1;
 
 mb_button_data[button] = action;
 return 0;
} 

//==============================================================================
// Associate a joystick switch with a key.
//
//   pass: char *key
// return: int				0 if no error, else -1
//==============================================================================
int joystick_kbjoy_key (char *key)
{
 int scan;
 char keyx[10];

 strncpy(keyx, key, sizeof(keyx));
 keyx[sizeof(keyx)-1] = 0;

 toupper_string(keyx, keyx);

 scan = string_search(key_to_6545, keyx);
 if (scan == -1)
    {
     scan = string_search(emu_commands, keyx);
     if (scan == -1)
        return -1;
     scan += 64;   
    }    

 kb_scan = scan;
  
 if (! kb_joy_init)
    {
     memset(kb_button_data_buf, 0, sizeof(kb_button_data_buf));
     kb_joy_init = 1;
    } 

 return 0;
} 

//==============================================================================
// Save a button for later use.
//
//   pass: int button
// return: int				0 if no error, else -1
//==============================================================================
int joystick_kbjoy_button (int button)
{
 if ((button < 0) || (button >= sizeof(kb_button_data_buf)))
    return -1;

 kb_button_data_buf[button] = kb_scan+1;
 return 0;
}

//==============================================================================
// Move the joystick key mappings to storage
//
//   pass: int set
//         char *set_str
// return: int				0 if no error, else -1
//==============================================================================
int joystick_kbjoy_set (int set, char *set_str)
{
 char set_chr;
 
 if (strlen(set_str) == 1)
    {
     set_chr = toupper(set_str[0]);
     if ((set_chr >= 'A') && (set_chr <= 'Z'))
        set = set_chr - 'A';
    }

 if ((set < 0) || (set >= JOY_KB_ENTRIES))
    return -1;

 memcpy(&kb_button_data[set], kb_button_data_buf, sizeof(kb_button_data_buf));
 kb_joy_init = 0;

 return 0;
} 

//==============================================================================
// List the key names available for the joystick key mappings.
//
//   pass: void
// return: void
//==============================================================================
void joystick_kbjoy_listkeys (void)
{
 int i = 0;
 
 while (i < 64)
    {
     xprintf("%-8s", key_to_6545[i++]);
     if ((i % 8) == 0)
        xprintf("\n");        
    }
} 

//==============================================================================
// List the command names available for the joystick command mappings.
//
//   pass: void
// return: void
//==============================================================================
void joystick_kbjoy_listcommands (void)
{
 int i = 0;
 
 while (emu_commands[i])
    {
     xprintf("%-16s", emu_commands[i++]);
     if ((i % 4) == 0)
        xprintf("\n");        
    }
 if ((i % 4) != 0)
    xprintf("\n");
} 

//==============================================================================
// Select a joystick key mappings set and enable joystick keys.
//
//   pass: int set
//         char *set_str
// return: int				0 if no error, else -1
//==============================================================================
int joystick_kbjoy_select (int set, char *set_str)
{
 char set_chr;
 
 if (strlen(set_str) == 1)
    {
     set_chr = toupper(set_str[0]);
     if ((set_chr >= 'A') && (set_chr <= 'Z'))
        set = set_chr - 'A';
    }

 if ((set < 0) || (set >= JOY_KB_ENTRIES))
    return -1;

 kb_button_data_use = &kb_button_data[set][0];
 joystick.kbd = 1;
 joystick.set = set;

 return 0;
} 

//==============================================================================
// joystick commands
//
//   pass: int cmd			video commands
//         int p                        parameter
// return: void
//==============================================================================
void joystick_command (int cmd, int p)
{
 switch (cmd)
    {
     case  EMU_CMD_JOYSTICK : if ((p >= 0) && (p < 26))
                                 joystick_kbjoy_select(p, "");
                              if (p == 26)
                                 {
                                  joystick.mbee = 0;
                                  joystick.kbd = 0;
                                 }
                              if (p == 27)
                                 joystick.mbee = 1;
                              break;   
    }
}
