//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                           Z80 debugging module                             *
//*                                                                            *
//*                   Copyright (C) 2007-2009 Stewart Kay                      *
//******************************************************************************
//
// Provides debugging tools to aid in the development of the emulator.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.2.0 - 19 July 2009, S.J.Kay
// - Added pio_regdump() to z80debug_command() function.
// - Changed debug.break_point to use one-time or repeated break points.
//
// v4.0.0 - 8 June 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Replaced all z80mem accesses with z80api_read_mem() function.
// - Fixed incorrect disassembly of IX and IY values by reloading op_1 and
//   op_2 in the z80debug_dis_Op1andC0_xC0_e_x05() function.  This bug was
//   likely to have been introduced when the original code was restructured.
// - Changed z80debug_before() to return 1 instead of 0 when step needed.
//
// v3.1.0 - 31 January 2009, S.J.Kay
// - Moved commands over from function.c to new z80debug_command() function.
// - Removed debug.debugcount = -1 from z80debug_init(), this and functions
//   z80debug_deinit() and z80debug_reset() were not even being called.
// - Changed all printf() calls to use a local xprintf() function.
// - Changed all putchar() to use local xprintf() function.
//
// v3.0.0 - 6 October 2008, S.J.Kay
// - SDL_Delay() replaced by time_delay_ms() function.
//
// v2.8.0 - 27 August 2008, S.J.Kay
// - Added code for pause mode, this uses methods similar to debug.
// - Added SDL_Delay() functions to step mode to reduce host CPU time.
//
// v2.7.0 - 22 May 2008, S.J.Kay
// - Added structure emu_t and moved variables into it.
//
// v2.3.0 - 23 January 2008, S.J.Kay
// - Created a new file to implement the Z80 debugging tools.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models. 
 *  Copyright (C) 2007-2009 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#include "ubee512.h"
#include "z80debug.h"
#include "z80api.h"
#include "z80.h"
#include "pio.h"
#include "crtc.h"
#include "rtc.h"
#include "function.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
z80regs_t z80before;
z80regs_t z80now;

debug_t debug;

static int dump_addr_x;

static char z80dis[300];

typedef char *str_p;

static int z80pc_before;
static int z80pc_same;

extern emu_t emu;
extern regdump_t regdump;

//==============================================================================
// z80debug initialise.
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_init (void)
{
 dump_addr_x = debug.dump_addr;
 return 0;
}

//==============================================================================
// z80debug de-initialise.
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_deinit (void)
{
 return 0;
}

//==============================================================================
// z80debug reset.
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_reset (void)
{
 z80pc_same = 0;
 return 0;
}

//==============================================================================
// z80debug disassembler.
//
// The original source code was by (c)1992 Markus Fritz.  I have made a lot
// of changes to it for this application. The source has been recoded to
// make it more compact and broken up into functional parts.
//
//   pass: void
// return: int 				0
//==============================================================================
#define mem16u op_1+(op_2 << 8)
#define memrel adr+2+(int8_t)z80api_read_mem(adr+1) 

 char t[100];

 uint8_t op_a;
 uint8_t op_b;
 uint8_t op_c;
 
 uint8_t op_1;
 uint8_t op_2;

 static str_p single_reg[] = {"b","c","d","e","h","l","(hl)","a"};
 static str_p double_reg[] = {"bc","de","hl","sp"};
 static str_p cond_flags[] = {"nz","z","nc","c","po","pe","p","m"};
 static str_p arith_ops[] = {"add~a,","adc~a,","sub~","sbc~a,","and~","xor~","or~","cp~"};

 static str_p str_a[] = {"rlca","rrca","rla","rra","daa","cpl","scf","ccf"};
 static str_p str_b[] = {"rlc","rrc","rl","rr","sla","sra","???","srl"};
 static str_p str_c[] = {"neg","???","???","???","???","???","???","???"}; 
 static str_p str_d[] = {"retn","reti","???","???","???","???","???","???"};
 static str_p str_e[] = {"ld~i,a","???","ld~a,i","???","rrd","rld","???","???"};
 static str_p str_f[] = {"ldi","cpi","ini","outi","???","???","???","???",
                             "ldd","cpd","ind","outd","???","???","???","???",
                             "ldir","cpir","inir","otir","???","???","???","???",
                             "lddr","cpdr","indr","otdr","???","???","???","???"
                        }; 
 static str_p str_g[] = {"rlc","rrc","rl","rr","sla","sra","???","srl"};
 
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x00 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x00 (int adr, str_p t)
{
 switch (op_c)
    {
     case 0x00: switch (op_b)
                   {
                    case 0x00: strcpy(t, "nop");
                               break;
                    case 0x01: strcpy(t, "ex~af,af'");
                               break;
                    case 0x02: sprintf(t, "djnz~%04xh", memrel);
                               break;
                    case 0x03: sprintf(t, "jr~%04xh", memrel);
                               break;
                      default: sprintf(t, "jr~%s,%04xh", cond_flags[op_b & B8(00000011)], memrel);
                               break;
                   }
                break;
     case 0x01: if (op_a & B8(00001000))
                   sprintf(t, "add~hl,%s", double_reg[op_b >> 1]);
                else
                   sprintf(t, "ld~%s,%04xh", double_reg[op_b >> 1], mem16u);
                break;
                               
     case 0x02: switch (op_b)
                   {
                    case 0x00: strcpy(t,"ld~(bc),a");
                               break;
                    case 0x01: strcpy(t,"ld~a,(bc)");
                               break;
                    case 0x02: strcpy(t,"ld~(de),a");
                               break;
                    case 0x03: strcpy(t,"ld~a,(de)");
                               break;
                    case 0x04: sprintf(t, "ld~(%04xh),hl", mem16u);
                               break;
                    case 0x05: sprintf(t, "ld~hl,(%04xh)", mem16u);
                               break;
                    case 0x06: sprintf(t, "ld~(%04xh),a", mem16u);
                               break;
                    case 0x07: sprintf(t, "ld~a,(%04xh)", mem16u);
                               break;
                   }
                break;

     case 0x03: if (op_a & B8(00001000))
                   sprintf(t, "dec~%s", double_reg[op_b >> 1]);
                else
                   sprintf(t, "inc~%s", double_reg[op_b >> 1]);                               
                break;
     case 0x04: sprintf(t, "inc~%s", single_reg[op_b]);
                break;
     case 0x05: sprintf(t, "dec~%s", single_reg[op_b]);
                break;
     case 0x06: sprintf(t, "ld~%s,%02xh", single_reg[op_b], op_1);
                break;
     case 0x07: strcpy(t, str_a[op_b]);
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x40 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x40 (int adr, str_p t)
{
 if (op_b == op_c)
    strcpy(t,"halt");
 else
    sprintf(t, "ld~%s,%s", single_reg[op_b], single_reg[op_c]);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x80 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x80 (int adr, str_p t)
{
 sprintf(t, "%s%s", arith_ops[op_b], single_reg[op_c]);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0_e_x03 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0_e_x03 (int adr, str_p t)
{
 switch (op_b)
    {
     case 0x00: sprintf(t, "jp~%04xh", mem16u);
                break;
     case 0x01: op_a = z80api_read_mem(++adr);
                op_b = (op_a >> 3) & B8(00000111);
                op_c = op_a & B8(00000111);
                switch (op_a & B8(11000000))
                   {
                    case 0x00: sprintf(t, "%s~%s", str_b[op_b], single_reg[op_c]);
                               break;
                    case 0x40: sprintf(t, "bit~%d,%s", op_b, single_reg[op_c]);
                               break;
                    case 0x80: sprintf(t, "res~%d,%s", op_b, single_reg[op_c]);
                               break;
                    case 0xC0: sprintf(t, "set~%d,%s", op_b, single_reg[op_c]);
                               break;
                   }
                break;
     case 0x02: sprintf(t, "out~(%02xh),a", op_1);
                break;
     case 0x03: sprintf(t, "in~a,(%02xh)", op_1);
                break;
     case 0x04: strcpy(t,"ex~(sp),hl");
                break;
     case 0x05: strcpy(t,"ex~de,hl");
                break;
     case 0x06: strcpy(t,"di");
                break;
     case 0x07: strcpy(t,"ei");
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0_e_x05 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0_e_x05 (int adr, str_p t)
{
 char index_reg[3];

 if (op_b & B8(00000001))
    {
     switch (op_b >> 1)
        {
         case 0x00: sprintf(t, "call~%04xh", mem16u);
                    break;
         case 0x02: op_a = z80api_read_mem(++adr);
                    op_b = (op_a >> 3) & B8(00000111);
                    op_c = op_a & B8(00000111);
                    switch (op_a & B8(11000000))
                       {
                        case 0x40: switch (op_c)
                                      {
                                       case 0x00: sprintf(t, "in~%s,(c)", single_reg[op_b]);
                                                  break;
                                       case 0x01: sprintf(t, "out~(c),%s", single_reg[op_b]);
                                                  break;
                                       case 0x02: if (op_b & B8(00000001))
                                                     sprintf(t, "adc~hl,%s", double_reg[op_b >> 1]);
                                                  else
                                                     sprintf(t, "sbc~hl,%s", double_reg[op_b >> 1]);                                                     
                                                  break;
                                       case 0x03: if (op_b & B8(00000001))
                                                     sprintf(t, "ld~%s,(%04xh)", double_reg[op_b >> 1], mem16u);
                                                  else
                                                     sprintf(t, "ld~(%04xh),%s", mem16u, double_reg[op_b >> 1]);
                                                  break;
                                       case 0x04: strcpy(t, str_c[op_b]);
                                                  break;
                                       case 0x05: strcpy(t, str_d[op_b]);
                                                  break;
                                       case 0x06: sprintf(t, "im~%d", op_b-1);
                                                  break;
                                       case 0x07: strcpy(t, str_e[op_b]);
                                                  break;
                                      }
                                   break;
                        case 0x80: strcpy(t, str_f[op_a & B8(00011111)]);
                                   break;
                       }
                    break;
           default: strcpy(index_reg, (op_a & B8(00100000))?"iy":"ix");
                    op_a = z80api_read_mem(++adr);
                    op_1 = z80api_read_mem(adr+1);
                    op_2 = z80api_read_mem(adr+2);
                    switch (op_a)
                       {
                        case 0x09: sprintf(t, "add~%s,bc", index_reg);
                                   break;
                        case 0x19: sprintf(t, "add~%s,de", index_reg);
                                   break;
                        case 0x21: sprintf(t, "ld~%s,%04xh", index_reg, mem16u);
                                   break;
                        case 0x22: sprintf(t, "ld~(%04xh),%s", mem16u, index_reg);
                                   break;
                        case 0x23: sprintf(t, "inc~%s", index_reg);
                                   break;
                        case 0x29: sprintf(t, "add~%s,%s", index_reg, index_reg);
                                   break;
                        case 0x2A: sprintf(t, "ld~%s,(%04xh)", index_reg, mem16u);
                                   break;
                        case 0x2B: sprintf(t, "dec~%s", index_reg);
                                   break;
                        case 0x34: sprintf(t, "inc~(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x35: sprintf(t, "dec~(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x36: sprintf(t, "ld~(%s+%02xh),%02xh", index_reg, op_1, op_2);
                                   break;
                        case 0x39: sprintf(t, "add~%s,sp", index_reg);
                                   break;
                        case 0x46:
                        case 0x4E:
                        case 0x56:
                        case 0x5E:
                        case 0x66:
                        case 0x6E: sprintf(t, "ld~%s,(%s+%02xh)", single_reg[(op_a >> 3) & B8(00000111)], index_reg, op_1);
                                   break;
                        case 0x70:
                        case 0x71:
                        case 0x72:
                        case 0x73:
                        case 0x74:
                        case 0x75:
                        case 0x77: sprintf(t, "ld~(%s+%02xh),%s", index_reg, op_1, single_reg[op_a & B8(00000111)]);
                                   break;
                        case 0x7E: sprintf(t, "ld~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x86: sprintf(t, "add~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x8E: sprintf(t, "adc~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x96: sprintf(t, "sub~,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x9E: sprintf(t, "sbc~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xA6: sprintf(t, "and~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xAE: sprintf(t, "xor~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xB6: sprintf(t, "or~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xBE: sprintf(t, "cp~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xE1: sprintf(t, "pop~%s", index_reg);
                                   break;
                        case 0xE3: sprintf(t, "ex~(sp),%s", index_reg);
                                   break;
                        case 0xE5: sprintf(t, "push~%s", index_reg);
                                   break;
                        case 0xE9: sprintf(t, "jp~(%s)", index_reg);
                                   break;
                        case 0xF9: sprintf(t, "ld~sp,%s", index_reg);
                                   break;
                        case 0xCB: op_a = op_2;
                                   op_b = (op_a >> 3) & B8(00000111);
                                   switch (op_a & B8(11000000))
                                      {
                                       case 0x00: sprintf(t, "%s~(%s+%02xh)", str_g[op_b], index_reg, op_1);
                                                  break;
                                       case 0x40: sprintf(t, "bit~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                       case 0x80: sprintf(t, "res~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                       case 0xC0: sprintf(t, "set~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                      }
                                   break;
                       }
                    break;
        }
    }
 else
    {
     if ((op_b >> 1) == B8(00000011))
        sprintf(t, "push~af");
     else
        sprintf(t, "push~%s", double_reg[op_b >> 1]);     
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0 (int adr, str_p t)
{
 switch (op_c)
    {
     case 0x00: sprintf(t, "ret~%s", cond_flags[op_b]);
                break;
     case 0x01: if (op_b & B8(00000001))
                   {
                    switch (op_b >> 1)
                       {
                        case 0x00: strcpy(t, "ret");
                                   break;
                        case 0x01: strcpy(t, "exx");
                                   break;
                        case 0x02: strcpy(t, "jp~(hl)");
                                   break;
                        case 0x03: strcpy(t, "ld~sp,hl");
                                   break;
                       }
                   }
                else
                   {
                    strcpy(t, "pop~");
                    if ((op_b >> 1) == B8(00000011))
                       strcat(t, "af");
                    else
                       strcat(t, double_reg[op_b >> 1]);
                   }
                break;
     case 0x02: sprintf(t, "jp~%s,%04xh", cond_flags[op_b], mem16u);
                break;
     case 0x03: z80debug_dis_Op1andC0_xC0_e_x03(adr, t);
                break;
     case 0x04: sprintf(t, "call~%s,%04xh", cond_flags[op_b], mem16u);
                break;
     case 0x05: z80debug_dis_Op1andC0_xC0_e_x05(adr, t);
                break;
     case 0x06: sprintf(t, "%s%02xh", arith_ops[op_b], op_1);
                break;
     case 0x07: sprintf(t, "rst~%02xh", op_a & B8(00111000));
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_disassemble (int adr, str_p dis)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_disassemble (int adr, str_p dis)
{
 static char t[40];
 
 op_a = z80api_read_mem(adr);
 op_b = (op_a >> 3) & B8(00000111);
 op_c = op_a & B8(00000111);

 op_1 = z80api_read_mem(adr+1);
 op_2 = z80api_read_mem(adr+2);

 switch (op_a & B8(11000000))
    {
     case 0x00: z80debug_dis_Op1andC0_x00(adr, t);
                break;
     case 0x40: z80debug_dis_Op1andC0_x40(adr, t);
                break;
     case 0x80: z80debug_dis_Op1andC0_x80(adr, t);
                break;
     case 0xC0: z80debug_dis_Op1andC0_xC0(adr, t);
                break;
    }

 sprintf(dis, "%04x:~%s", (uint16_t)adr, t);    
}

//==============================================================================
// z80debug before instruction execution.
//
// Dissasemble an instruction at current Z80 PC.  This is placed into a string
// and is output in the z80debug_after() function.
//
//   pass: void
// return: int				step wanted if 1 or -1 if none
//==============================================================================
int z80debug_before (void)
{
 int bp = 0;
 
 z80api_get_regs(&z80before);

 // if emulator is paused (no code execution)
 if (emu.paused)
    {
     if (emu.quit || emu.reset)
        emu.z80_blocks = 0;     // kill the current block loop
     else
        time_delay_ms(1);  // prevent excessive host CPU time whilst paused
     return -1;
    } 

 // test and do the disassembly only if needed
 if ((debug.mode == 2) || ((debug.mode == 1) && (debug.step)))
    z80debug_disassemble(z80before.pc, z80dis);

 // if step mode and nothing to do then return (no code execution)
 if ((debug.mode == 1) && (! debug.step))
    {
     if (emu.quit || emu.reset)
        emu.z80_blocks = 0;     // kill the current block loop
     else
        time_delay_ms(1);  // prevent excessive host CPU time while in step mode
     return -1;
    } 

 // check for code break points in code execution
 if (debug.mode != 1)	// if not in step mode
    {
     if (debug.break_point[z80before.pc])
        {
         xprintf("Z80 Debugging break point at PC: 0x%04x\n", z80before.pc);
         if (debug.break_point[z80before.pc] == 1)
            debug.break_point[z80before.pc] = 0;
         bp = 1;
        }

     if ((debug.debugcount == debug.break_point_count) && debug.debugcount)
        {
         xprintf("Z80 Debugging break point at count: 0x%08x (%d) at PC: 0x%04x\n",
         debug.debugcount, debug.debugcount, z80before.pc);
         bp = 1;
        }
     
     if (bp)
        {
         debug.mode = 1;  // change to step mode
         debug.step = 0;
         return -1;
        }
    }

 z80pc_before = z80before.pc;
 z80pc_same++;
 debug.debugcount++;

 return 1;
}

//==============================================================================
// z80debug after instruction execution.
//
// Show state of registers.  The registers and other information displayed
// depends on option settings.
//
//   pass: void
// return: void
//==============================================================================
void z80debug_after (void)
{
 int x;

 int i;
 int count;
 int begin;
 int l;

 int SP_C;
 
 static char dis[20];
 static char flags[9];

 if ((debug.mode == 2) || ((debug.mode == 1) && (debug.step)))
    {
     z80api_get_regs(&z80now);
     
     // don't keep showing same address unless bigger than max BC value
     if ((z80pc_before == z80now.pc) && (z80pc_same < 66000L))
        return;

     z80pc_same = 0;   
     
     if (debug.step)
        debug.step--;

     // show the instruction counter if enabled
     if (debug.showcount)
         xprintf("%08x ", debug.debugcount);

     // break the disassembled code up for columned output
     l = strlen(z80dis);
     i = 0;

     for (x = 0; x < 3; x++)
        {
         count = 0;
         begin = i;
         dis[0] = 0;
         while ((i < l) && (z80dis[i] != '~'))
            dis[count++] = z80dis[i++];
         dis[count] = 0;
         i++;	 // skip pass the ~ character for next iteration
         if (x < 2)
            xprintf("%-8s", dis);
         else   
            xprintf("%-16s", dis);        
        }

     // show the status of the common registers     
     flags[0] = z80now.af & B8(00000001)?'C':'-';
     flags[1] = z80now.af & B8(00000010)?'N':'-';     
     flags[2] = z80now.af & B8(00000100)?'P':'-';
     flags[3] = z80now.af & B8(00001000)?'-':'-';        
     flags[4] = z80now.af & B8(00010000)?'A':'-';
     flags[5] = z80now.af & B8(00100000)?'-':'-';
     flags[6] = z80now.af & B8(01000000)?'Z':'-';
     flags[7] = z80now.af & B8(10000000)?'S':'-';
     flags[8] = 0;
     xprintf("FLG: %s AF:%04x BC:%04x DE:%04x HL:%04x", flags, z80now.af, z80now.bc, z80now.de, z80now.hl);

     // show the index and SP registers
     if (debug.show & B8(00000001))
        {
         SP_C = (z80api_read_mem(z80now.sp+1) << 8) | z80api_read_mem(z80now.sp); // fetch the previous contents
         xprintf(" IX:%04x IY:%04x SP:%04x (%04x)\n", z80now.ix, z80now.iy, z80now.sp, SP_C);
        }
     else
        xprintf("\n");     

     // show the alternate and I and R registers
     if (debug.show & B8(00000010))
        {
         if (debug.showcount)
            xprintf("%57s", "Alternate ");
         else   
            xprintf("%48s", "Alternate ");
         xprintf("AF:%04x BC:%04x DE:%04x HL:%04x I:%02x R:%02x\n",
         z80now.af_p, z80now.bc_p, z80now.de_p, z80now.hl_p, z80now.i, z80now.r);
        }
    }
}

//==============================================================================
// Dump Z80 memory.
//
//   pass: int addr		Z80 memory address
//         int lines            number of lines (16 bytes per line), 0=64K
//         int header           header type, 0=none
// return: void
//==============================================================================
void z80debug_dump (int addr, int lines, int htype)
{
 int c;
 int a;
 char x;

 if ((lines > 4096) || (lines <= 0))
    lines = 4096;			// 64K maximium

 if (htype)
    {
     xprintf("     ");
     for (c = 0; c < 16; c++)
        xprintf(" %02X", c);
     xprintf("  ");
     for (c = 0; c < 16; c++)
        xprintf("%1X", c);
     xprintf("\n");
    } 

 while (lines--)
    {
     addr &= 0xFFFF;
     a = addr;
     
     xprintf("%04X:", a);
     for (c = 0; c < 16; c++)
        {
         a &= 0xFFFF;
         xprintf(" %02X", z80api_read_mem(a));
         a++;
        } 
     xprintf(" ");        
     
     xprintf("\"");
     a = addr;
     for (c = 0; c < 16; c++)
        {
         a &= 0xFFFF;
         x = z80api_read_mem(a);
         a++;
         if (isalpha(x) || isdigit(x))
            xprintf("%c", x);
         else
            xprintf(".");
        }
     xprintf("\"\n");
     
     addr += 16;
    }
}    

//==============================================================================
// z80debug commands.
//
// These may be called from various locations including options so any values
// here must not be initialised in z80debug_init() or z80debug_reset()
// functions.
//
//   pass: int cmd			debugging command
// return: void
//==============================================================================
void z80debug_command (int cmd)
{
 switch (cmd)
    {
     case      EMU_CMD_DUMP : dump_addr_x = debug.dump_addr;
                              xprintf("\n");     
                              z80debug_dump(dump_addr_x, 8, 1);
                              break;
     case   EMU_CMD_DUMP_N1 : dump_addr_x += 0x0080;
                              xprintf("\n");
                              z80debug_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_N2 : dump_addr_x += 0x1000;
                              xprintf("\n");
                              z80debug_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_B1 : dump_addr_x -= 0x0080;
                              xprintf("\n");
                              z80debug_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_B2 : dump_addr_x -= 0x1000;
                              xprintf("\n");
                              z80debug_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case  EMU_CMD_DUMP_REP : xprintf("\n");
                              z80debug_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case  EMU_CMD_DUMPREGS : if (regdump.crtc)
                                 crtc_regdump();
                              if (regdump.rtc)
                                 rtc_regdump();
                              if (regdump.z80)
                                 z80api_regdump();
                              if (regdump.pio)
                                 pio_regdump();
                              break;
     case    EMU_CMD_DBGOFF : if (! emu.paused)
                                 {
                                  debug.debug = 0;
                                  debug.mode = 0;
                                  debug.step = 0;
                                  emu.z80_blocks = 0; // restart normal execution
                                  xprintf("Debug mode is now off\n");
                                 } 
                              break;
     case     EMU_CMD_DBGON : if (! emu.paused)
                                 {
                                  debug.debug = 1;
                                  debug.mode = 0;
                                  debug.step = 0;
                                  xprintf("Debug mode is now on\n");
                                 } 
                              break;
     case  EMU_CMD_DBGTRACE : if (! emu.paused)
                                 {
                                  if (debug.mode == 2)
                                     debug.mode = 0;
                                  else
                                     {   
                                      debug.debug = 1;
                                      debug.mode = 2;
                                     }
                                 }    
                              break;
     case EMU_CMD_DBGSTEP01 : if (! emu.paused)
                                 debug.step = 1;
                              break;
     case EMU_CMD_DBGSTEP10 : if (! emu.paused)
                                 debug.step = 10;
                              break;                               
     case EMU_CMD_DBGSTEP20 : if (! emu.paused)
                                 debug.step = 20;
                              break;
     case     EMU_CMD_PAUSE : if (emu.paused)
                                 {
                                  debug.debug = 0;
                                  emu.paused = 0;
                                  emu.z80_blocks = 0;
                                 }
                              else
                                 {
                                  debug.debug = 1;
                                  debug.mode = 0;
                                  debug.step = 0;                                     
                                  emu.paused = 1;
                                 }
                              break;
    }

 if ((cmd >= EMU_CMD_DBGSTEP01) && (cmd <= EMU_CMD_DBGSTEP20))
    {
     if (debug.step > 1)
        xprintf("\n");
     debug.debug = 1;
     debug.mode = 1;
    }
}
